# encoding: utf-8

from utils import get_pos_list
from reportsdata.salechallengereport import SaleChallengeGenerator, SaleChallengePresenter
from salechallengereport.repository import SaleChallengeRepository, OrderRepository, ProductRepository


class ReportGenerator(object):

    @staticmethod
    def generate(mb_context, _, operator_id, sale_challenge_id):
        operator_id = str(int(0 if operator_id in (None, 'None') else operator_id))
        if operator_id == '0':
            operator_id = None

        order_repository = OrderRepository(mb_context, get_pos_list(mb_context))
        sale_challenge_repository = SaleChallengeRepository(mb_context)
        product_repository = ProductRepository(mb_context)

        generator = SaleChallengeGenerator(
                order_repository,
                product_repository,
                sale_challenge_repository,
                sale_challenge_id,
                operator_id
        )
        presenter = SaleChallengePresenter(38)

        return generator, presenter




def check_and_remove(order_list, order, state_type, state_value, match=True):
    if order.order_id not in order_list:
        return

    if not order.items or (state_match(order, state_type, state_value) and match):
        order_list.remove(order.order_id)


def state_match(order, state_type, state_value):
    if state_type.lower() == "order_state":
        state_from_history = next((s for s in order.state_history if s.state == state_value), None)
        return (state_from_history.state == state_value) if state_from_history else False

    return order.prod_state == state_value



# -*- coding: utf-8 -*-
"""
    werkzeug.test
    ~~~~~~~~~~~~~

    This module implements a client to WSGI applications for testing.

    :copyright: 2007 Pallets
    :license: BSD-3-Clause
"""
import mimetypes
import sys
from io import BytesIO
from itertools import chain
from random import random
from tempfile import TemporaryFile
from time import time

from ._compat import iteritems
from ._compat import iterlists
from ._compat import itervalues
from ._compat import make_literal_wrapper
from ._compat import reraise
from ._compat import string_types
from ._compat import text_type
from ._compat import to_bytes
from ._compat import wsgi_encoding_dance
from ._internal import _get_environ
from .datastructures import CallbackDict
from .datastructures import CombinedMultiDict
from .datastructures import EnvironHeaders
from .datastructures import FileMultiDict
from .datastructures import Headers
from .datastructures import MultiDict
from .http import dump_cookie
from .http import dump_options_header
from .http import parse_options_header
from .urls import iri_to_uri
from .urls import url_encode
from .urls import url_fix
from .urls import url_parse
from .urls import url_unparse
from .urls import url_unquote
from .utils import get_content_type
from .wrappers import BaseRequest
from .wsgi import ClosingIterator
from .wsgi import get_current_url

try:
    from urllib.request import Request as U2Request
except ImportError:
    from urllib2 import Request as U2Request

try:
    from http.cookiejar import CookieJar
except ImportError:
    from cookielib import CookieJar


def stream_encode_multipart(
    values, use_tempfile=True, threshold=1024 * 500, boundary=None, charset="utf-8"
):
    """Encode a dict of values (either strings or file descriptors or
    :class:`FileStorage` objects.) into a multipart encoded string stored
    in a file descriptor.
    """
    if boundary is None:
        boundary = "---------------WerkzeugFormPart_%s%s" % (time(), random())
    _closure = [BytesIO(), 0, False]

    if use_tempfile:

        def write_binary(string):
            stream, total_length, on_disk = _closure
            if on_disk:
                stream.write(string)
            else:
                length = len(string)
                if length + _closure[1] <= threshold:
                    stream.write(string)
                else:
                    new_stream = TemporaryFile("wb+")
                    new_stream.write(stream.getvalue())
                    new_stream.write(string)
                    _closure[0] = new_stream
                    _closure[2] = True
                _closure[1] = total_length + length

    else:
        write_binary = _closure[0].write

    def write(string):
        write_binary(string.encode(charset))

    if not isinstance(values, MultiDict):
        values = MultiDict(values)

    for key, values in iterlists(values):
        for value in values:
            write('--%s\r\nContent-Disposition: form-data; name="%s"' % (boundary, key))
            reader = getattr(value, "read", None)
            if reader is not None:
                filename = getattr(value, "filename", getattr(value, "name", None))
                content_type = getattr(value, "content_type", None)
                if content_type is None:
                    content_type = (
                        filename
                        and mimetypes.guess_type(filename)[0]
                        or "application/octet-stream"
                    )
                if filename is not None:
                    write('; filename="%s"\r\n' % filename)
                else:
                    write("\r\n")
                write("Content-Type: %s\r\n\r\n" % content_type)
                while 1:
                    chunk = reader(16384)
                    if not chunk:
                        break
                    write_binary(chunk)
            else:
                if not isinstance(value, string_types):
                    value = str(value)

                value = to_bytes(value, charset)
                write("\r\n\r\n")
                write_binary(value)
            write("\r\n")
    write("--%s--\r\n" % boundary)

    length = int(_closure[0].tell())
    _closure[0].seek(0)
    return _closure[0], length, boundary


def encode_multipart(values, boundary=None, charset="utf-8"):
    """Like `stream_encode_multipart` but returns a tuple in the form
    (``boundary``, ``data``) where data is a bytestring.
    """
    stream, length, boundary = stream_encode_multipart(
        values, use_tempfile=False, boundary=boundary, charset=charset
    )
    return boundary, stream.read()


class _TestCookieHeaders(object):

    """A headers adapter for cookielib
    """

    def __init__(self, headers):
        self.headers = headers

    def getheaders(self, name):
        headers = []
        name = name.lower()
        for k, v in self.headers:
            if k.lower() == name:
                headers.append(v)
        return headers

    def get_all(self, name, default=None):
        rv = []
        for k, v in self.headers:
            if k.lower() == name.lower():
                rv.append(v)
        return rv or default or []


class _TestCookieResponse(object):

    """Something that looks like a httplib.HTTPResponse, but is actually just an
    adapter for our test responses to make them available for cookielib.
    """

    def __init__(self, headers):
        self.headers = _TestCookieHeaders(headers)

    def info(self):
        return self.headers


class _TestCookieJar(CookieJar):

    """A cookielib.CookieJar modified to inject and read cookie headers from
    and to wsgi environments, and wsgi application responses.
    """

    def inject_wsgi(self, environ):
        """Inject the cookies as client headers into the server's wsgi
        environment.
        """
        cvals = ["%s=%s" % (c.name, c.value) for c in self]

        if cvals:
            environ["HTTP_COOKIE"] = "; ".join(cvals)
        else:
            environ.pop("HTTP_COOKIE", None)

    def extract_wsgi(self, environ, headers):
        """Extract the server's set-cookie headers as cookies into the
        cookie jar.
        """
        self.extract_cookies(
            _TestCookieResponse(headers), U2Request(get_current_url(environ))
        )


def _iter_data(data):
    """Iterates over a `dict` or :class:`MultiDict` yielding all keys and
    values.
    This is used to iterate over the data passed to the
    :class:`EnvironBuilder`.
    """
    if isinstance(data, MultiDict):
        for key, values in iterlists(data):
            for value in values:
                yield key, value
    else:
        for key, values in iteritems(data):
            if isinstance(values, list):
                for value in values:
                    yield key, value
            else:
                yield key, values


class EnvironBuilder(object):
    """This class can be used to conveniently create a WSGI environment
    for testing purposes.  It can be used to quickly create WSGI environments
    or request objects from arbitrary data.

    The signature of this class is also used in some other places as of
    Werkzeug 0.5 (:func:`create_environ`, :meth:`BaseResponse.from_values`,
    :meth:`Client.open`).  Because of this most of the functionality is
    available through the constructor alone.

    Files and regular form data can be manipulated independently of each
    other with the :attr:`form` and :attr:`files` attributes, but are
    passed with the same argument to the constructor: `data`.

    `data` can be any of these values:

    -   a `str` or `bytes` object: The object is converted into an
        :attr:`input_stream`, the :attr:`content_length` is set and you have to
        provide a :attr:`content_type`.
    -   a `dict` or :class:`MultiDict`: The keys have to be strings. The values
        have to be either any of the following objects, or a list of any of the
        following objects:

        -   a :class:`file`-like object:  These are converted into
            :class:`FileStorage` objects automatically.
        -   a `tuple`:  The :meth:`~FileMultiDict.add_file` method is called
            with the key and the unpacked `tuple` items as positional
            arguments.
        -   a `str`:  The string is set as form data for the associated key.
    -   a file-like object: The object content is loaded in memory and then
        handled like a regular `str` or a `bytes`.

    :param path: the path of the request.  In the WSGI environment this will
                 end up as `PATH_INFO`.  If the `query_string` is not defined
                 and there is a question mark in the `path` everything after
                 it is used as query string.
    :param base_url: the base URL is a URL that is used to extract the WSGI
                     URL scheme, host (server name + server port) and the
                     script root (`SCRIPT_NAME`).
    :param query_string: an optional string or dict with URL parameters.
    :param method: the HTTP method to use, defaults to `GET`.
    :param input_stream: an optional input stream.  Do not specify this and
                         `data`.  As soon as an input stream is set you can't
                         modify :attr:`args` and :attr:`files` unless you
                         set the :attr:`input_stream` to `None` again.
    :param content_type: The content type for the request.  As of 0.5 you
                         don't have to provide this when specifying files
                         and form data via `data`.
    :param content_length: The content length for the request.  You don't
                           have to specify this when providing data via
                           `data`.
    :param errors_stream: an optional error stream that is used for
                          `wsgi.errors`.  Defaults to :data:`stderr`.
    :param multithread: controls `wsgi.multithread`.  Defaults to `False`.
    :param multiprocess: controls `wsgi.multiprocess`.  Defaults to `False`.
    :param run_once: controls `wsgi.run_once`.  Defaults to `False`.
    :param headers: an optional list or :class:`Headers` object of headers.
    :param data: a string or dict of form data or a file-object.
                 See explanation above.
    :param json: An object to be serialized and assigned to ``data``.
        Defaults the content type to ``"application/json"``.
        Serialized with the function assigned to :attr:`json_dumps`.
    :param environ_base: an optional dict of environment defaults.
    :param environ_overrides: an optional dict of environment overrides.
    :param charset: the charset used to encode unicode data.

    .. versionadded:: 0.15
        The ``json`` param and :meth:`json_dumps` method.

    .. versionadded:: 0.15
        The environ has keys ``REQUEST_URI`` and ``RAW_URI`` containing
        the path before perecent-decoding. This is not part of the WSGI
        PEP, but many WSGI servers include it.

    .. versionchanged:: 0.6
       ``path`` and ``base_url`` can now be unicode strings that are
       encoded with :func:`iri_to_uri`.
    """

    #: the server protocol to use.  defaults to HTTP/1.1
    server_protocol = "HTTP/1.1"

    #: the wsgi version to use.  defaults to (1, 0)
    wsgi_version = (1, 0)

    #: the default request class for :meth:`get_request`
    request_class = BaseRequest

    import json

    #: The serialization function used when ``json`` is passed.
    json_dumps = staticmethod(json.dumps)
    del json

    def __init__(
        self,
        path="/",
        base_url=None,
        query_string=None,
        method="GET",
        input_stream=None,
        content_type=None,
        content_length=None,
        errors_stream=None,
        multithread=False,
        multiprocess=False,
        run_once=False,
        headers=None,
        data=None,
        environ_base=None,
        environ_overrides=None,
        charset="utf-8",
        mimetype=None,
        json=None,
    ):
        path_s = make_literal_wrapper(path)
        if query_string is not None and path_s("?") in path:
            raise ValueError("Query string is defined in the path and as an argument")
        if query_string is None and path_s("?") in path:
            path, query_string = path.split(path_s("?"), 1)
        self.charset = charset
        self.path = iri_to_uri(path)
        if base_url is not None:
            base_url = url_fix(iri_to_uri(base_url, charset), charset)
        self.base_url = base_url
        if isinstance(query_string, (bytes, text_type)):
            self.query_string = query_string
        else:
            if query_string is None:
                query_string = MultiDict()
            elif not isinstance(query_string, MultiDict):
                query_string = MultiDict(query_string)
            self.args = query_string
        self.method = method
        if headers is None:
            headers = Headers()
        elif not isinstance(headers, Headers):
            headers = Headers(headers)
        self.headers = headers
        if content_type is not None:
            self.content_type = content_type
        if errors_stream is None:
            errors_stream = sys.stderr
        self.errors_stream = errors_stream
        self.multithread = multithread
        self.multiprocess = multiprocess
        self.run_once = run_once
        self.environ_base = environ_base
        self.environ_overrides = environ_overrides
        self.input_stream = input_stream
        self.content_length = content_length
        self.closed = False

        if json is not None:
            if data is not None:
                raise TypeError("can't provide both json and data")

            data = self.json_dumps(json)

            if self.content_type is None:
                self.content_type = "application/json"

        if data:
            if input_stream is not None:
                raise TypeError("can't provide input stream and data")
            if hasattr(data, "read"):
                data = data.read()
            if isinstance(data, text_type):
                data = data.encode(self.charset)
            if isinstance(data, bytes):
                self.input_stream = BytesIO(data)
                if self.content_length is None:
                    self.content_length = len(data)
            else:
                for key, value in _iter_data(data):
                    if isinstance(value, (tuple, dict)) or hasattr(value, "read"):
                        self._add_file_from_data(key, value)
                    else:
                        self.form.setlistdefault(key).append(value)

        if mimetype is not None:
            self.mimetype = mimetype

    @classmethod
    def from_environ(cls, environ, **kwargs):
        """Turn an environ dict back into a builder. Any extra kwargs
        override the args extracted from the environ.

        .. versionadded:: 0.15
        """
        headers = Headers(EnvironHeaders(environ))
        out = {
            "path": environ["PATH_INFO"],
            "base_url": cls._make_base_url(
                environ["wsgi.url_scheme"], headers.pop("Host"), environ["SCRIPT_NAME"]
            ),
            "query_string": environ["QUERY_STRING"],
            "method": environ["REQUEST_METHOD"],
            "input_stream": environ["wsgi.input"],
            "content_type": headers.pop("Content-Type", None),
            "content_length": headers.pop("Content-Length", None),
            "errors_stream": environ["wsgi.errors"],
            "multithread": environ["wsgi.multithread"],
            "multiprocess": environ["wsgi.multiprocess"],
            "run_once": environ["wsgi.run_once"],
            "headers": headers,
        }
        out.update(kwargs)
        return cls(**out)

    def _add_file_from_data(self, key, value):
        """Called in the EnvironBuilder to add files from the data dict."""
        if isinstance(value, tuple):
            self.files.add_file(key, *value)
        else:
            self.files.add_file(key, value)

    @staticmethod
    def _make_base_url(scheme, host, script_root):
        return url_unparse((scheme, host, script_root, "", "")).rstrip("/") + "/"

    @property
    def base_url(self):
        """The base URL is used to extract the URL scheme, host name,
        port, and root path.
        """
        return self._make_base_url(self.url_scheme, self.host, self.script_root)

    @base_url.setter
    def base_url(self, value):
        if value is None:
            scheme = "http"
            netloc = "localhost"
            script_root = ""
        else:
            scheme, netloc, script_root, qs, anchor = url_parse(value)
            if qs or anchor:
                raise ValueError("base url must not contain a query string or fragment")
        self.script_root = script_root.rstrip("/")
        self.host = netloc
        self.url_scheme = scheme

    @property
    def content_type(self):
        """The content type for the request.  Reflected from and to
        the :attr:`headers`.  Do not set if you set :attr:`files` or
        :attr:`form` for auto detection.
        """
        ct = self.headers.get("Content-Type")
        if ct is None and not self._input_stream:
            if self._files:
                return "multipart/form-data"
            if self._form:
                return "application/x-www-form-urlencoded"
            return None
        return ct

    @content_type.setter
    def content_type(self, value):
        if value is None:
            self.headers.pop("Content-Type", None)
        else:
            self.headers["Content-Type"] = value

    @property
    def mimetype(self):
        """The mimetype (content type without charset etc.)

        .. versionadded:: 0.14
        """
        ct = self.content_type
        return ct.split(";")[0].strip() if ct else None

    @mimetype.setter
    def mimetype(self, value):
        self.content_type = get_content_type(value, self.charset)

    @property
    def mimetype_params(self):
        """ The mimetype parameters as dict.  For example if the
        content type is ``text/html; charset=utf-8`` the params would be
        ``{'charset': 'utf-8'}``.

        .. versionadded:: 0.14
        """

        def on_update(d):
            self.headers["Content-Type"] = dump_options_header(self.mimetype, d)

        d = parse_options_header(self.headers.get("content-type", ""))[1]
        return CallbackDict(d, on_update)

    @property
    def content_length(self):
        """The content length as integer.  Reflected from and to the
        :attr:`headers`.  Do not set if you set :attr:`files` or
        :attr:`form` for auto detection.
        """
        return self.headers.get("Content-Length", type=int)

    @content_length.setter
    def content_length(self, value):
        if value is None:
            self.headers.pop("Content-Length", None)
        else:
            self.headers["Content-Length"] = str(value)

    def _get_form(self, name, storage):
        """Common behavior for getting the :attr:`form` and
        :attr:`files` properties.

        :param name: Name of the internal cached attribute.
        :param storage: Storage class used for the data.
        """
        if self.input_stream is not None:
            raise AttributeError("an input stream is defined")

        rv = getattr(self, name)

        if rv is None:
            rv = storage()
            setattr(self, name, rv)

        return rv

    def _set_form(self, name, value):
        """Common behavior for setting the :attr:`form` and
        :attr:`files` properties.

        :param name: Name of the internal cached attribute.
        :param value: Value to assign to the attribute.
        """
        self._input_stream = None
        setattr(self, name, value)

    @property
    def form(self):
        """A :class:`MultiDict` of form values."""
        return self._get_form("_form", MultiDict)

    @form.setter
    def form(self, value):
        self._set_form("_form", value)

    @property
    def files(self):
        """A :class:`FileMultiDict` of uploaded files. Use
        :meth:`~FileMultiDict.add_file` to add new files.
        """
        return self._get_form("_files", FileMultiDict)

    @files.setter
    def files(self, value):
        self._set_form("_files", value)

    @property
    def input_stream(self):
        """An optional input stream.  If you set this it will clear
        :attr:`form` and :attr:`files`.
        """
        return self._input_stream

    @input_stream.setter
    def input_stream(self, value):
        self._input_stream = value
        self._form = None
        self._files = None

    @property
    def query_string(self):
        """The query string.  If you set this to a string
        :attr:`args` will no longer be available.
        """
        if self._query_string is None:
            if self._args is not None:
                return url_encode(self._args, charset=self.charset)
            return ""
        return self._query_string

    @query_string.setter
    def query_string(self, value):
        self._query_string = value
        self._args = None

    @property
    def args(self):
        """The URL arguments as :class:`MultiDict`."""
        if self._query_string is not None:
            raise AttributeError("a query string is defined")
        if self._args is None:
            self._args = MultiDict()
        return self._args

    @args.setter
    def args(self, value):
        self._query_string = None
        self._args = value

    @property
    def server_name(self):
        """The server name (read-only, use :attr:`host` to set)"""
        return self.host.split(":", 1)[0]

    @property
    def server_port(self):
        """The server port as integer (read-only, use :attr:`host` to set)"""
        pieces = self.host.split(":", 1)
        if len(pieces) == 2 and pieces[1].isdigit():
            return int(pieces[1])
        if self.url_scheme == "https":
            return 443
        return 80

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def close(self):
        """Closes all files.  If you put real :class:`file` objects into the
        :attr:`files` dict you can call this method to automatically close
        them all in one go.
        """
        if self.closed:
            return
        try:
            files = itervalues(self.files)
        except AttributeError:
            files = ()
        for f in files:
            try:
                f.close()
            except Exception:
                pass
        self.closed = True

    def get_environ(self):
        """Return the built environ.

        .. versionchanged:: 0.15
            The content type and length headers are set based on
            input stream detection. Previously this only set the WSGI
            keys.
        """
        input_stream = self.input_stream
        content_length = self.content_length

        mimetype = self.mimetype
        content_type = self.content_type

        if input_stream is not None:
            start_pos = input_stream.tell()
            input_stream.seek(0, 2)
            end_pos = input_stream.tell()
            input_stream.seek(start_pos)
            content_length = end_pos - start_pos
        elif mimetype == "multipart/form-data":
            values = CombinedMultiDict([self.form, self.files])
            input_stream, content_length, boundary = stream_encode_multipart(
                values, charset=self.charset
            )
            content_type = mimetype + '; boundary="%s"' % boundary
        elif mimetype == "application/x-www-form-urlencoded":
            # XXX: py2v3 review
            values = url_encode(self.form, charset=self.charset)
            values = values.encode("ascii")
            content_length = len(values)
            input_stream = BytesIO(values)
        else:
            input_stream = BytesIO()

        result = {}
        if self.environ_base:
            result.update(self.environ_base)

        def _path_encode(x):
            return wsgi_encoding_dance(url_unquote(x, self.charset), self.charset)

        qs = wsgi_encoding_dance(self.query_string)

        result.update(
            {
                "REQUEST_METHOD": self.method,
                "SCRIPT_NAME": _path_encode(self.script_root),
                "PATH_INFO": _path_encode(self.path),
                "QUERY_STRING": qs,
                # Non-standard, added by mod_wsgi, uWSGI
                "REQUEST_URI": wsgi_encoding_dance(self.path),
                # Non-standard, added by gunicorn
                "RAW_URI": wsgi_encoding_dance(self.path),
                "SERVER_NAME": self.server_name,
                "SERVER_PORT": str(self.server_port),
                "HTTP_HOST": self.host,
                "SERVER_PROTOCOL": self.server_protocol,
                "wsgi.version": self.wsgi_version,
                "wsgi.url_scheme": self.url_scheme,
                "wsgi.input": input_stream,
                "wsgi.errors": self.errors_stream,
                "wsgi.multithread": self.multithread,
                "wsgi.multiprocess": self.multiprocess,
                "wsgi.run_once": self.run_once,
            }
        )

        headers = self.headers.copy()

        if content_type is not None:
            result["CONTENT_TYPE"] = content_type
            headers.set("Content-Type", content_type)

        if content_length is not None:
            result["CONTENT_LENGTH"] = str(content_length)
            headers.set("Content-Length", content_length)

        for key, value in headers.to_wsgi_list():
            result["HTTP_%s" % key.upper().replace("-", "_")] = value

        if self.environ_overrides:
            result.update(self.environ_overrides)

        return result

    def get_request(self, cls=None):
        """Returns a request with the data.  If the request class is not
        specified :attr:`request_class` is used.

        :param cls: The request wrapper to use.
        """
        if cls is None:
            cls = self.request_class
        return cls(self.get_environ())


class ClientRedirectError(Exception):
    """If a redirect loop is detected when using follow_redirects=True with
    the :cls:`Client`, then this exception is raised.
    """


class Client(object):
    """This class allows you to send requests to a wrapped application.

    The response wrapper can be a class or factory function that takes
    three arguments: app_iter, status and headers.  The default response
    wrapper just returns a tuple.

    Example::

        class ClientResponse(BaseResponse):
            ...

        client = Client(MyApplication(), response_wrapper=ClientResponse)

    The use_cookies parameter indicates whether cookies should be stored and
    sent for subsequent requests. This is True by default, but passing False
    will disable this behaviour.

    If you want to request some subdomain of your application you may set
    `allow_subdomain_redirects` to `True` as if not no external redirects
    are allowed.

    .. versionadded:: 0.5
       `use_cookies` is new in this version.  Older versions did not provide
       builtin cookie support.

    .. versionadded:: 0.14
       The `mimetype` parameter was added.

    .. versionadded:: 0.15
        The ``json`` parameter.
    """

    def __init__(
        self,
        application,
        response_wrapper=None,
        use_cookies=True,
        allow_subdomain_redirects=False,
    ):
        self.application = application
        self.response_wrapper = response_wrapper
        if use_cookies:
            self.cookie_jar = _TestCookieJar()
        else:
            self.cookie_jar = None
        self.allow_subdomain_redirects = allow_subdomain_redirects

    def set_cookie(
        self,
        server_name,
        key,
        value="",
        max_age=None,
        expires=None,
        path="/",
        domain=None,
        secure=None,
        httponly=False,
        samesite=None,
        charset="utf-8",
    ):
        """Sets a cookie in the client's cookie jar.  The server name
        is required and has to match the one that is also passed to
        the open call.
        """
        assert self.cookie_jar is not None, "cookies disabled"
        header = dump_cookie(
            key,
            value,
            max_age,
            expires,
            path,
            domain,
            secure,
            httponly,
            charset,
            samesite=samesite,
        )
        environ = create_environ(path, base_url="http://" + server_name)
        headers = [("Set-Cookie", header)]
        self.cookie_jar.extract_wsgi(environ, headers)

    def delete_cookie(self, server_name, key, path="/", domain=None):
        """Deletes a cookie in the test client."""
        self.set_cookie(
            server_name, key, expires=0, max_age=0, path=path, domain=domain
        )

    def run_wsgi_app(self, environ, buffered=False):
        """Runs the wrapped WSGI app with the given environment."""
        if self.cookie_jar is not None:
            self.cookie_jar.inject_wsgi(environ)
        rv = run_wsgi_app(self.application, environ, buffered=buffered)
        if self.cookie_jar is not None:
            self.cookie_jar.extract_wsgi(environ, rv[2])
        return rv

    def resolve_redirect(self, response, new_location, environ, buffered=False):
        """Perform a new request to the location given by the redirect
        response to the previous request.
        """
        scheme, netloc, path, qs, anchor = url_parse(new_location)
        builder = EnvironBuilder.from_environ(environ, query_string=qs)

        to_name_parts = netloc.split(":", 1)[0].split(".")
        from_name_parts = builder.server_name.split(".")

        if to_name_parts != [""]:
            # The new location has a host, use it for the base URL.
            builder.url_scheme = scheme
            builder.host = netloc
        else:
            # A local redirect with autocorrect_location_header=False
            # doesn't have a host, so use the request's host.
            to_name_parts = from_name_parts

        # Explain why a redirect to a different server name won't be followed.
        if to_name_parts != from_name_parts:
            if to_name_parts[-len(from_name_parts) :] == from_name_parts:
                if not self.allow_subdomain_redirects:
                    raise RuntimeError("Following subdomain redirects is not enabled.")
            else:
                raise RuntimeError("Following external redirects is not supported.")

        path_parts = path.split("/")
        root_parts = builder.script_root.split("/")

        if path_parts[: len(root_parts)] == root_parts:
            # Strip the script root from the path.
            builder.path = path[len(builder.script_root) :]
        else:
            # The new location is not under the script root, so use the
            # whole path and clear the previous root.
            builder.path = path
            builder.script_root = ""

        status_code = int(response[1].split(None, 1)[0])

        # Only 307 and 308 preserve all of the original request.
        if status_code not in {307, 308}:
            # HEAD is preserved, everything else becomes GET.
            if builder.method != "HEAD":
                builder.method = "GET"

            # Clear the body and the headers that describe it.
            builder.input_stream = None
            builder.content_type = None
            builder.content_length = None
            builder.headers.pop("Transfer-Encoding", None)

        # Disable the response wrapper while handling redirects. Not
        # thread safe, but the client should not be shared anyway.
        old_response_wrapper = self.response_wrapper
        self.response_wrapper = None

        try:
            return self.open(builder, as_tuple=True, buffered=buffered)
        finally:
            self.response_wrapper = old_response_wrapper

    def open(self, *args, **kwargs):
        """Takes the same arguments as the :class:`EnvironBuilder` class with
        some additions:  You can provide a :class:`EnvironBuilder` or a WSGI
        environment as only argument instead of the :class:`EnvironBuilder`
        arguments and two optional keyword arguments (`as_tuple`, `buffered`)
        that change the type of the return value or the way the application is
        executed.

        .. versionchanged:: 0.5
           If a dict is provided as file in the dict for the `data` parameter
           the content type has to be called `content_type` now instead of
           `mimetype`.  This change was made for consistency with
           :class:`werkzeug.FileWrapper`.

            The `follow_redirects` parameter was added to :func:`open`.

        Additional parameters:

        :param as_tuple: Returns a tuple in the form ``(environ, result)``
        :param buffered: Set this to True to buffer the application run.
                         This will automatically close the application for
                         you as well.
        :param follow_redirects: Set this to True if the `Client` should
                                 follow HTTP redirects.
        """
        as_tuple = kwargs.pop("as_tuple", False)
        buffered = kwargs.pop("buffered", False)
        follow_redirects = kwargs.pop("follow_redirects", False)
        environ = None
        if not kwargs and len(args) == 1:
            if isinstance(args[0], EnvironBuilder):
                environ = args[0].get_environ()
            elif isinstance(args[0], dict):
                environ = args[0]
        if environ is None:
            builder = EnvironBuilder(*args, **kwargs)
            try:
                environ = builder.get_environ()
            finally:
                builder.close()

        response = self.run_wsgi_app(environ.copy(), buffered=buffered)

        # handle redirects
        redirect_chain = []
        while 1:
            status_code = int(response[1].split(None, 1)[0])
            if (
                status_code not in {301, 302, 303, 305, 307, 308}
                or not follow_redirects
            ):
                break

            # Exhaust intermediate response bodies to ensure middleware
            # that returns an iterator runs any cleanup code.
            if not buffered:
                for _ in response[0]:
                    pass

            new_location = response[2]["location"]
            new_redirect_entry = (new_location, status_code)
            if new_redirect_entry in redirect_chain:
                raise ClientRedirectError("loop detected")
            redirect_chain.append(new_redirect_entry)
            environ, response = self.resolve_redirect(
                response, new_location, environ, buffered=buffered
            )

        if self.response_wrapper is not None:
            response = self.response_wrapper(*response)
        if as_tuple:
            return environ, response
        return response

    def get(self, *args, **kw):
        """Like open but method is enforced to GET."""
        kw["method"] = "GET"
        return self.open(*args, **kw)

    def patch(self, *args, **kw):
        """Like open but method is enforced to PATCH."""
        kw["method"] = "PATCH"
        return self.open(*args, **kw)

    def post(self, *args, **kw):
        """Like open but method is enforced to POST."""
        kw["method"] = "POST"
        return self.open(*args, **kw)

    def head(self, *args, **kw):
        """Like open but method is enforced to HEAD."""
        kw["method"] = "HEAD"
        return self.open(*args, **kw)

    def put(self, *args, **kw):
        """Like open but method is enforced to PUT."""
        kw["method"] = "PUT"
        return self.open(*args, **kw)

    def delete(self, *args, **kw):
        """Like open but method is enforced to DELETE."""
        kw["method"] = "DELETE"
        return self.open(*args, **kw)

    def options(self, *args, **kw):
        """Like open but method is enforced to OPTIONS."""
        kw["method"] = "OPTIONS"
        return self.open(*args, **kw)

    def trace(self, *args, **kw):
        """Like open but method is enforced to TRACE."""
        kw["method"] = "TRACE"
        return self.open(*args, **kw)

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.application)


def create_environ(*args, **kwargs):
    """Create a new WSGI environ dict based on the values passed.  The first
    parameter should be the path of the request which defaults to '/'.  The
    second one can either be an absolute path (in that case the host is
    localhost:80) or a full path to the request with scheme, netloc port and
    the path to the script.

    This accepts the same arguments as the :class:`EnvironBuilder`
    constructor.

    .. versionchanged:: 0.5
       This function is now a thin wrapper over :class:`EnvironBuilder` which
       was added in 0.5.  The `headers`, `environ_base`, `environ_overrides`
       and `charset` parameters were added.
    """
    builder = EnvironBuilder(*args, **kwargs)
    try:
        return builder.get_environ()
    finally:
        builder.close()


def run_wsgi_app(app, environ, buffered=False):
    """Return a tuple in the form (app_iter, status, headers) of the
    application output.  This works best if you pass it an application that
    returns an iterator all the time.

    Sometimes applications may use the `write()` callable returned
    by the `start_response` function.  This tries to resolve such edge
    cases automatically.  But if you don't get the expected output you
    should set `buffered` to `True` which enforces buffering.

    If passed an invalid WSGI application the behavior of this function is
    undefined.  Never pass non-conforming WSGI applications to this function.

    :param app: the application to execute.
    :param buffered: set to `True` to enforce buffering.
    :return: tuple in the form ``(app_iter, status, headers)``
    """
    environ = _get_environ(environ)
    response = []
    buffer = []

    def start_response(status, headers, exc_info=None):
        if exc_info is not None:
            reraise(*exc_info)
        response[:] = [status, headers]
        return buffer.append

    app_rv = app(environ, start_response)
    close_func = getattr(app_rv, "close", None)
    app_iter = iter(app_rv)

    # when buffering we emit the close call early and convert the
    # application iterator into a regular list
    if buffered:
        try:
            app_iter = list(app_iter)
        finally:
            if close_func is not None:
                close_func()

    # otherwise we iterate the application iter until we have a response, chain
    # the already received data with the already collected data and wrap it in
    # a new `ClosingIterator` if we need to restore a `close` callable from the
    # original return value.
    else:
        for item in app_iter:
            buffer.append(item)
            if response:
                break
        if buffer:
            app_iter = chain(buffer, app_iter)
        if close_func is not None and app_iter is not app_rv:
            app_iter = ClosingIterator(app_iter, close_func)

    return app_iter, response[0], Headers(response[1])



import json

from ._Generator import Generator  # noqa
from ._Presenter import Presenter  # noqa
from ._ReportJSONEncoder import ReportJSONEncoder


class Reporter(object):
    def __init__(self, generator, presenter):
        # type: (Generator, Presenter) -> None
        self.generator = generator
        self.presenter = presenter

    def create_report(self):
        dto = self.generator.generate_data()
        report = self.presenter.present(dto)
        return json.dumps(report.get_parts(), cls=ReportJSONEncoder, encoding="utf-8")



"""
Python 3 reorganized the standard library (PEP 3108). This module exposes
several standard library modules to Python 2 under their new Python 3
names.

It is designed to be used as follows::

    from future import standard_library
    standard_library.install_aliases()

And then these normal Py3 imports work on both Py3 and Py2::

    import builtins
    import copyreg
    import queue
    import reprlib
    import socketserver
    import winreg    # on Windows only
    import test.support
    import html, html.parser, html.entites
    import http, http.client, http.server
    import http.cookies, http.cookiejar
    import urllib.parse, urllib.request, urllib.response, urllib.error, urllib.robotparser
    import xmlrpc.client, xmlrpc.server

    import _thread
    import _dummy_thread
    import _markupbase

    from itertools import filterfalse, zip_longest
    from sys import intern
    from collections import UserDict, UserList, UserString
    from collections import OrderedDict, Counter     # even on Py2.6
    from subprocess import getoutput, getstatusoutput
    from subprocess import check_output              # even on Py2.6

(The renamed modules and functions are still available under their old
names on Python 2.)

This is a cleaner alternative to this idiom (see
http://docs.pythonsprints.com/python3_porting/py-porting.html)::

    try:
        import queue
    except ImportError:
        import Queue as queue


Limitations
-----------
We don't currently support these modules, but would like to::

    import dbm
    import dbm.dumb
    import dbm.gnu
    import collections.abc  # on Py33
    import pickle     # should (optionally) bring in cPickle on Python 2

"""

from __future__ import absolute_import, division, print_function

import sys
import logging
import imp
import contextlib
import types
import copy
import os

# Make a dedicated logger; leave the root logger to be configured
# by the application.
flog = logging.getLogger('future_stdlib')
_formatter = logging.Formatter(logging.BASIC_FORMAT)
_handler = logging.StreamHandler()
_handler.setFormatter(_formatter)
flog.addHandler(_handler)
flog.setLevel(logging.WARN)

from future.utils import PY2, PY3

# The modules that are defined under the same names on Py3 but with
# different contents in a significant way (e.g. submodules) are:
#   pickle (fast one)
#   dbm
#   urllib
#   test
#   email

REPLACED_MODULES = set(['test', 'urllib', 'pickle', 'dbm'])  # add email and dbm when we support it

# The following module names are not present in Python 2.x, so they cause no
# potential clashes between the old and new names:
#   http
#   html
#   tkinter
#   xmlrpc
# Keys: Py2 / real module names
# Values: Py3 / simulated module names
RENAMES = {
           # 'cStringIO': 'io',  # there's a new io module in Python 2.6
                                 # that provides StringIO and BytesIO
           # 'StringIO': 'io',   # ditto
           # 'cPickle': 'pickle',
           '__builtin__': 'builtins',
           'copy_reg': 'copyreg',
           'Queue': 'queue',
           'future.moves.socketserver': 'socketserver',
           'ConfigParser': 'configparser',
           'repr': 'reprlib',
           # 'FileDialog': 'tkinter.filedialog',
           # 'tkFileDialog': 'tkinter.filedialog',
           # 'SimpleDialog': 'tkinter.simpledialog',
           # 'tkSimpleDialog': 'tkinter.simpledialog',
           # 'tkColorChooser': 'tkinter.colorchooser',
           # 'tkCommonDialog': 'tkinter.commondialog',
           # 'Dialog': 'tkinter.dialog',
           # 'Tkdnd': 'tkinter.dnd',
           # 'tkFont': 'tkinter.font',
           # 'tkMessageBox': 'tkinter.messagebox',
           # 'ScrolledText': 'tkinter.scrolledtext',
           # 'Tkconstants': 'tkinter.constants',
           # 'Tix': 'tkinter.tix',
           # 'ttk': 'tkinter.ttk',
           # 'Tkinter': 'tkinter',
           '_winreg': 'winreg',
           'thread': '_thread',
           'dummy_thread': '_dummy_thread',
           # 'anydbm': 'dbm',   # causes infinite import loop
           # 'whichdb': 'dbm',  # causes infinite import loop
           # anydbm and whichdb are handled by fix_imports2
           # 'dbhash': 'dbm.bsd',
           # 'dumbdbm': 'dbm.dumb',
           # 'dbm': 'dbm.ndbm',
           # 'gdbm': 'dbm.gnu',
           'future.moves.xmlrpc': 'xmlrpc',
           # 'future.backports.email': 'email',    # for use by urllib
           # 'DocXMLRPCServer': 'xmlrpc.server',
           # 'SimpleXMLRPCServer': 'xmlrpc.server',
           # 'httplib': 'http.client',
           # 'htmlentitydefs' : 'html.entities',
           # 'HTMLParser' : 'html.parser',
           # 'Cookie': 'http.cookies',
           # 'cookielib': 'http.cookiejar',
           # 'BaseHTTPServer': 'http.server',
           # 'SimpleHTTPServer': 'http.server',
           # 'CGIHTTPServer': 'http.server',
           # 'future.backports.test': 'test',  # primarily for renaming test_support to support
           # 'commands': 'subprocess',
           # 'urlparse' : 'urllib.parse',
           # 'robotparser' : 'urllib.robotparser',
           # 'abc': 'collections.abc',   # for Py33
           # 'future.utils.six.moves.html': 'html',
           # 'future.utils.six.moves.http': 'http',
           'future.moves.html': 'html',
           'future.moves.http': 'http',
           # 'future.backports.urllib': 'urllib',
           # 'future.utils.six.moves.urllib': 'urllib',
           'future.moves._markupbase': '_markupbase',
          }


# It is complicated and apparently brittle to mess around with the
# ``sys.modules`` cache in order to support "import urllib" meaning two
# different things (Py2.7 urllib and backported Py3.3-like urllib) in different
# contexts. So we require explicit imports for these modules.
assert len(set(RENAMES.values()) & set(REPLACED_MODULES)) == 0


# Harmless renames that we can insert.
# These modules need names from elsewhere being added to them:
#   subprocess: should provide getoutput and other fns from commands
#               module but these fns are missing: getstatus, mk2arg,
#               mkarg
#   re:         needs an ASCII constant that works compatibly with Py3

# etc: see lib2to3/fixes/fix_imports.py

# (New module name, new object name, old module name, old object name)
MOVES = [('collections', 'UserList', 'UserList', 'UserList'),
         ('collections', 'UserDict', 'UserDict', 'UserDict'),
         ('collections', 'UserString','UserString', 'UserString'),
         ('itertools', 'filterfalse','itertools', 'ifilterfalse'),
         ('itertools', 'zip_longest','itertools', 'izip_longest'),
         ('sys', 'intern','__builtin__', 'intern'),
         # The re module has no ASCII flag in Py2, but this is the default.
         # Set re.ASCII to a zero constant. stat.ST_MODE just happens to be one
         # (and it exists on Py2.6+).
         ('re', 'ASCII','stat', 'ST_MODE'),
         ('base64', 'encodebytes','base64', 'encodestring'),
         ('base64', 'decodebytes','base64', 'decodestring'),
         ('subprocess', 'getoutput', 'commands', 'getoutput'),
         ('subprocess', 'getstatusoutput', 'commands', 'getstatusoutput'),
         ('subprocess', 'check_output', 'future.backports.misc', 'check_output'),
         ('math', 'ceil', 'future.backports.misc', 'ceil'),
         ('collections', 'OrderedDict', 'future.backports.misc', 'OrderedDict'),
         ('collections', 'Counter', 'future.backports.misc', 'Counter'),
         ('itertools', 'count', 'future.backports.misc', 'count'),
         ('reprlib', 'recursive_repr', 'future.backports.misc', 'recursive_repr'),
         ('functools', 'cmp_to_key', 'future.backports.misc', 'cmp_to_key'),

# This is no use, since "import urllib.request" etc. still fails:
#          ('urllib', 'error', 'future.moves.urllib', 'error'),
#          ('urllib', 'parse', 'future.moves.urllib', 'parse'),
#          ('urllib', 'request', 'future.moves.urllib', 'request'),
#          ('urllib', 'response', 'future.moves.urllib', 'response'),
#          ('urllib', 'robotparser', 'future.moves.urllib', 'robotparser'),
        ]


# A minimal example of an import hook:
# class WarnOnImport(object):
#     def __init__(self, *args):
#         self.module_names = args
#
#     def find_module(self, fullname, path=None):
#         if fullname in self.module_names:
#             self.path = path
#             return self
#         return None
#
#     def load_module(self, name):
#         if name in sys.modules:
#             return sys.modules[name]
#         module_info = imp.find_module(name, self.path)
#         module = imp.load_module(name, *module_info)
#         sys.modules[name] = module
#         flog.warning("Imported deprecated module %s", name)
#         return module


class RenameImport(object):
    """
    A class for import hooks mapping Py3 module names etc. to the Py2 equivalents.
    """
    # Different RenameImport classes are created when importing this module from
    # different source files. This causes isinstance(hook, RenameImport) checks
    # to produce inconsistent results. We add this RENAMER attribute here so
    # remove_hooks() and install_hooks() can find instances of these classes
    # easily:
    RENAMER = True

    def __init__(self, old_to_new):
        '''
        Pass in a dictionary-like object mapping from old names to new
        names. E.g. {'ConfigParser': 'configparser', 'cPickle': 'pickle'}
        '''
        self.old_to_new = old_to_new
        both = set(old_to_new.keys()) & set(old_to_new.values())
        assert (len(both) == 0 and
                len(set(old_to_new.values())) == len(old_to_new.values())), \
               'Ambiguity in renaming (handler not implemented)'
        self.new_to_old = dict((new, old) for (old, new) in old_to_new.items())

    def find_module(self, fullname, path=None):
        # Handles hierarchical importing: package.module.module2
        new_base_names = set([s.split('.')[0] for s in self.new_to_old])
        # Before v0.12: Was: if fullname in set(self.old_to_new) | new_base_names:
        if fullname in new_base_names:
            return self
        return None

    def load_module(self, name):
        path = None
        if name in sys.modules:
            return sys.modules[name]
        elif name in self.new_to_old:
            # New name. Look up the corresponding old (Py2) name:
            oldname = self.new_to_old[name]
            module = self._find_and_load_module(oldname)
            # module.__future_module__ = True
        else:
            module = self._find_and_load_module(name)
        # In any case, make it available under the requested (Py3) name
        sys.modules[name] = module
        return module

    def _find_and_load_module(self, name, path=None):
        """
        Finds and loads it. But if there's a . in the name, handles it
        properly.
        """
        bits = name.split('.')
        while len(bits) > 1:
            # Treat the first bit as a package
            packagename = bits.pop(0)
            package = self._find_and_load_module(packagename, path)
            try:
                path = package.__path__
            except AttributeError:
                # This could be e.g. moves.
                flog.debug('Package {0} has no __path__.'.format(package))
                if name in sys.modules:
                    return sys.modules[name]
                flog.debug('What to do here?')

        name = bits[0]
        module_info = imp.find_module(name, path)
        return imp.load_module(name, *module_info)


class hooks(object):
    """
    Acts as a context manager. Saves the state of sys.modules and restores it
    after the 'with' block.

    Use like this:

    >>> from future import standard_library
    >>> with standard_library.hooks():
    ...     import http.client
    >>> import requests

    For this to work, http.client will be scrubbed from sys.modules after the
    'with' block. That way the modules imported in the 'with' block will
    continue to be accessible in the current namespace but not from any
    imported modules (like requests).
    """
    def __enter__(self):
        # flog.debug('Entering hooks context manager')
        self.old_sys_modules = copy.copy(sys.modules)
        self.hooks_were_installed = detect_hooks()
        # self.scrubbed = scrub_py2_sys_modules()
        install_hooks()
        return self

    def __exit__(self, *args):
        # flog.debug('Exiting hooks context manager')
        # restore_sys_modules(self.scrubbed)
        if not self.hooks_were_installed:
            remove_hooks()
        # scrub_future_sys_modules()

# Sanity check for is_py2_stdlib_module(): We aren't replacing any
# builtin modules names:
if PY2:
    assert len(set(RENAMES.values()) & set(sys.builtin_module_names)) == 0


def is_py2_stdlib_module(m):
    """
    Tries to infer whether the module m is from the Python 2 standard library.
    This may not be reliable on all systems.
    """
    if PY3:
        return False
    if not 'stdlib_path' in is_py2_stdlib_module.__dict__:
        stdlib_files = [contextlib.__file__, os.__file__, copy.__file__]
        stdlib_paths = [os.path.split(f)[0] for f in stdlib_files]
        if not len(set(stdlib_paths)) == 1:
            # This seems to happen on travis-ci.org. Very strange. We'll try to
            # ignore it.
            flog.warn('Multiple locations found for the Python standard '
                         'library: %s' % stdlib_paths)
        # Choose the first one arbitrarily
        is_py2_stdlib_module.stdlib_path = stdlib_paths[0]

    if m.__name__ in sys.builtin_module_names:
        return True

    if hasattr(m, '__file__'):
        modpath = os.path.split(m.__file__)
        if (modpath[0].startswith(is_py2_stdlib_module.stdlib_path) and
            'site-packages' not in modpath[0]):
            return True

    return False


def scrub_py2_sys_modules():
    """
    Removes any Python 2 standard library modules from ``sys.modules`` that
    would interfere with Py3-style imports using import hooks. Examples are
    modules with the same names (like urllib or email).

    (Note that currently import hooks are disabled for modules like these
    with ambiguous names anyway ...)
    """
    if PY3:
        return {}
    scrubbed = {}
    for modulename in REPLACED_MODULES & set(RENAMES.keys()):
        if not modulename in sys.modules:
            continue

        module = sys.modules[modulename]

        if is_py2_stdlib_module(module):
            flog.debug('Deleting (Py2) {} from sys.modules'.format(modulename))
            scrubbed[modulename] = sys.modules[modulename]
            del sys.modules[modulename]
    return scrubbed


def scrub_future_sys_modules():
    """
    Deprecated.
    """
    return {}    

class suspend_hooks(object):
    """
    Acts as a context manager. Use like this:

    >>> from future import standard_library
    >>> standard_library.install_hooks()
    >>> import http.client
    >>> # ...
    >>> with standard_library.suspend_hooks():
    >>>     import requests     # incompatible with ``future``'s standard library hooks

    If the hooks were disabled before the context, they are not installed when
    the context is left.
    """
    def __enter__(self):
        self.hooks_were_installed = detect_hooks()
        remove_hooks()
        # self.scrubbed = scrub_future_sys_modules()
        return self

    def __exit__(self, *args):
        if self.hooks_were_installed:
            install_hooks()
        # restore_sys_modules(self.scrubbed)


def restore_sys_modules(scrubbed):
    """
    Add any previously scrubbed modules back to the sys.modules cache,
    but only if it's safe to do so.
    """
    clash = set(sys.modules) & set(scrubbed)
    if len(clash) != 0:
        # If several, choose one arbitrarily to raise an exception about
        first = list(clash)[0]
        raise ImportError('future module {} clashes with Py2 module'
                          .format(first))
    sys.modules.update(scrubbed)


def install_aliases():
    """
    Monkey-patches the standard library in Py2.6/7 to provide
    aliases for better Py3 compatibility.
    """
    if PY3:
        return
    # if hasattr(install_aliases, 'run_already'):
    #     return
    for (newmodname, newobjname, oldmodname, oldobjname) in MOVES:
        __import__(newmodname)
        # We look up the module in sys.modules because __import__ just returns the
        # top-level package:
        newmod = sys.modules[newmodname]
        # newmod.__future_module__ = True

        __import__(oldmodname)
        oldmod = sys.modules[oldmodname]

        obj = getattr(oldmod, oldobjname)
        setattr(newmod, newobjname, obj)

    # Hack for urllib so it appears to have the same structure on Py2 as on Py3
    import urllib
    from future.backports.urllib import request
    from future.backports.urllib import response
    from future.backports.urllib import parse
    from future.backports.urllib import error
    from future.backports.urllib import robotparser
    urllib.request = request
    urllib.response = response
    urllib.parse = parse
    urllib.error = error
    urllib.robotparser = robotparser
    sys.modules['urllib.request'] = request
    sys.modules['urllib.response'] = response
    sys.modules['urllib.parse'] = parse
    sys.modules['urllib.error'] = error
    sys.modules['urllib.robotparser'] = robotparser

    # Patch the test module so it appears to have the same structure on Py2 as on Py3
    try:
        import test
    except ImportError:
        pass
    try:
        from future.moves.test import support
    except ImportError:
        pass
    else:
        test.support = support
        sys.modules['test.support'] = support

    # Patch the dbm module so it appears to have the same structure on Py2 as on Py3
    try:
        import dbm
    except ImportError:
        pass
    else:
        from future.moves.dbm import dumb
        dbm.dumb = dumb
        sys.modules['dbm.dumb'] = dumb
        try:
            from future.moves.dbm import gnu
        except ImportError:
            pass
        else:
            dbm.gnu = gnu
            sys.modules['dbm.gnu'] = gnu
        try:
            from future.moves.dbm import ndbm
        except ImportError:
            pass
        else:
            dbm.ndbm = ndbm
            sys.modules['dbm.ndbm'] = ndbm

    # install_aliases.run_already = True


def install_hooks():
    """
    This function installs the future.standard_library import hook into
    sys.meta_path.
    """
    if PY3:
        return

    install_aliases()

    flog.debug('sys.meta_path was: {0}'.format(sys.meta_path))
    flog.debug('Installing hooks ...')

    # Add it unless it's there already
    newhook = RenameImport(RENAMES)
    if not detect_hooks():
        sys.meta_path.append(newhook)
    flog.debug('sys.meta_path is now: {0}'.format(sys.meta_path))


def enable_hooks():
    """
    Deprecated. Use install_hooks() instead. This will be removed by
    ``future`` v1.0.
    """
    install_hooks()


def remove_hooks(scrub_sys_modules=False):
    """
    This function removes the import hook from sys.meta_path.
    """
    if PY3:
        return
    flog.debug('Uninstalling hooks ...')
    # Loop backwards, so deleting items keeps the ordering:
    for i, hook in list(enumerate(sys.meta_path))[::-1]:
        if hasattr(hook, 'RENAMER'):
            del sys.meta_path[i]

    # Explicit is better than implicit. In the future the interface should
    # probably change so that scrubbing the import hooks requires a separate
    # function call. Left as is for now for backward compatibility with
    # v0.11.x.
    if scrub_sys_modules:
        scrub_future_sys_modules()


def disable_hooks():
    """
    Deprecated. Use remove_hooks() instead. This will be removed by
    ``future`` v1.0.
    """
    remove_hooks()


def detect_hooks():
    """
    Returns True if the import hooks are installed, False if not.
    """
    flog.debug('Detecting hooks ...')
    present = any([hasattr(hook, 'RENAMER') for hook in sys.meta_path])
    if present:
        flog.debug('Detected.')
    else:
        flog.debug('Not detected.')
    return present


# As of v0.12, this no longer happens implicitly:
# if not PY3:
#     install_hooks()


if not hasattr(sys, 'py2_modules'):
    sys.py2_modules = {}

def cache_py2_modules():
    """
    Currently this function is unneeded, as we are not attempting to provide import hooks
    for modules with ambiguous names: email, urllib, pickle.
    """
    if len(sys.py2_modules) != 0:
        return
    assert not detect_hooks()
    import urllib
    sys.py2_modules['urllib'] = urllib

    import email
    sys.py2_modules['email'] = email

    import pickle
    sys.py2_modules['pickle'] = pickle

    # Not all Python installations have test module. (Anaconda doesn't, for example.)
    # try:
    #     import test
    # except ImportError:
    #     sys.py2_modules['test'] = None
    # sys.py2_modules['test'] = test

    # import dbm
    # sys.py2_modules['dbm'] = dbm


def import_(module_name, backport=False):
    """
    Pass a (potentially dotted) module name of a Python 3 standard library
    module. This function imports the module compatibly on Py2 and Py3 and
    returns the top-level module.

    Example use:
        >>> http = import_('http.client')
        >>> http = import_('http.server')
        >>> urllib = import_('urllib.request')

    Then:
        >>> conn = http.client.HTTPConnection(...)
        >>> response = urllib.request.urlopen('http://mywebsite.com')
        >>> # etc.

    Use as follows:
        >>> package_name = import_(module_name)

    On Py3, equivalent to this:

        >>> import module_name

    On Py2, equivalent to this if backport=False:

        >>> from future.moves import module_name

    or to this if backport=True:

        >>> from future.backports import module_name

    except that it also handles dotted module names such as ``http.client``
    The effect then is like this:

        >>> from future.backports import module
        >>> from future.backports.module import submodule
        >>> module.submodule = submodule

    Note that this would be a SyntaxError in Python:

        >>> from future.backports import http.client

    """
    # Python 2.6 doesn't have importlib in the stdlib, so it requires
    # the backported ``importlib`` package from PyPI as a dependency to use
    # this function:
    import importlib

    if PY3:
        return __import__(module_name)
    else:
        # client.blah = blah
        # Then http.client = client
        # etc.
        if backport:
            prefix = 'future.backports'
        else:
            prefix = 'future.moves'
        parts = prefix.split('.') + module_name.split('.')

        modules = []
        for i, part in enumerate(parts):
            sofar = '.'.join(parts[:i+1])
            modules.append(importlib.import_module(sofar))
        for i, part in reversed(list(enumerate(parts))):
            if i == 0:
                break
            setattr(modules[i-1], part, modules[i])

        # Return the next-most top-level module after future.backports / future.moves:
        return modules[2]


def from_import(module_name, *symbol_names, **kwargs):
    """
    Example use:
        >>> HTTPConnection = from_import('http.client', 'HTTPConnection')
        >>> HTTPServer = from_import('http.server', 'HTTPServer')
        >>> urlopen, urlparse = from_import('urllib.request', 'urlopen', 'urlparse')

    Equivalent to this on Py3:

        >>> from module_name import symbol_names[0], symbol_names[1], ...

    and this on Py2:

        >>> from future.moves.module_name import symbol_names[0], ...

    or:

        >>> from future.backports.module_name import symbol_names[0], ...

    except that it also handles dotted module names such as ``http.client``.
    """

    if PY3:
        return __import__(module_name)
    else:
        if 'backport' in kwargs and bool(kwargs['backport']):
            prefix = 'future.backports'
        else:
            prefix = 'future.moves'
        parts = prefix.split('.') + module_name.split('.')
        module = importlib.import_module(prefix + '.' + module_name)
        output = [getattr(module, name) for name in symbol_names]
        if len(output) == 1:
            return output[0]
        else:
            return output


class exclude_local_folder_imports(object):
    """
    A context-manager that prevents standard library modules like configparser
    from being imported from the local python-future source folder on Py3.

    (This was need prior to v0.16.0 because the presence of a configparser
    folder would otherwise have prevented setuptools from running on Py3. Maybe
    it's not needed any more?)
    """
    def __init__(self, *args):
        assert len(args) > 0
        self.module_names = args
        # Disallow dotted module names like http.client:
        if any(['.' in m for m in self.module_names]):
            raise NotImplementedError('Dotted module names are not supported')

    def __enter__(self):
        self.old_sys_path = copy.copy(sys.path)
        self.old_sys_modules = copy.copy(sys.modules)
        if sys.version_info[0] < 3:
            return
        # The presence of all these indicates we've found our source folder,
        # because `builtins` won't have been installed in site-packages by setup.py:
        FUTURE_SOURCE_SUBFOLDERS = ['future', 'past', 'libfuturize', 'libpasteurize', 'builtins']

        # Look for the future source folder:
        for folder in self.old_sys_path:
            if all([os.path.exists(os.path.join(folder, subfolder))
                    for subfolder in FUTURE_SOURCE_SUBFOLDERS]):
                # Found it. Remove it.
                sys.path.remove(folder)

        # Ensure we import the system module:
        for m in self.module_names:
            # Delete the module and any submodules from sys.modules:
            # for key in list(sys.modules):
            #     if key == m or key.startswith(m + '.'):
            #         try:
            #             del sys.modules[key]
            #         except KeyError:
            #             pass
            try:
                module = __import__(m, level=0)
            except ImportError:
                # There's a problem importing the system module. E.g. the
                # winreg module is not available except on Windows.
                pass

    def __exit__(self, *args):
        # Restore sys.path and sys.modules:
        sys.path = self.old_sys_path
        for m in set(self.old_sys_modules.keys()) - set(sys.modules.keys()):
            sys.modules[m] = self.old_sys_modules[m]

TOP_LEVEL_MODULES = ['builtins',
                     'copyreg',
                     'html',
                     'http',
                     'queue',
                     'reprlib',
                     'socketserver',
                     'test',
                     'tkinter',
                     'winreg',
                     'xmlrpc',
                     '_dummy_thread',
                     '_markupbase',
                     '_thread',
                    ]

def import_top_level_modules():
    with exclude_local_folder_imports(*TOP_LEVEL_MODULES):
        for m in TOP_LEVEL_MODULES:
            try:
                __import__(m)
            except ImportError:     # e.g. winreg
                pass



from abc import ABCMeta, abstractmethod


class TableReleaser(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def release_all(self, pos_id):
        # type: (str) -> None
        raise NotImplementedError()



from logging import Logger

from application.model.benefit import ActionResult
from posot import OrderTaker

from messagebus import MessageBus
from typing import Dict

from application.model.order import Order
from application.model.transactions import AddBenefitToApplyUnitOfWork
from application.model.benefit import ActionTypes, Action
from application.model.conditions import Restrictions

from ._DiscountService import DiscountService
from ._OrderService import OrderService


class ActionService(object):
    def __init__(
        self,
        order_service,      # type: OrderService
        product_groups,     # type: Dict
        message_bus,        # type: MessageBus
        logger,             # type: Logger
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.order_service = order_service
        self.product_groups = product_groups
        self.logger = logger

    def apply_action(self, action, burnt_items, uow, pos_ot, restriction, on_total):
        # type: (Action, Dict, AddBenefitToApplyUnitOfWork, OrderTaker, Restrictions, bool) -> ActionResult

        order = self.order_service.get_order(uow.pos_id, order_picture=uow.order.order_picture)
        discount_ids = Order.get_available_discount_ids(pos_ot)
        service = DiscountService(
            pos_ot=pos_ot,
            product_groups=self.product_groups,
            message_bus=self.message_bus,
            logger=self.logger,
        )
        result = ActionResult([], {}, [], {}, False, False)

        if order.order_state == "TOTALED":
            if action.name == ActionTypes.discountAmountOnTotal.value:
                result = service.apply_discount_amount_on_total(action, order, restriction, discount_ids)

            if action.name == ActionTypes.discountPercentageOnTotal.value:
                result = service.apply_discount_percentage_on_total(
                    action=action,
                    order=order,
                    restriction=restriction,
                    available_discount_ids=discount_ids,
                    benefit=uow.benefit,
                )

            if action.name == ActionTypes.discountAmountByItems.value:
                result = service.apply_discount_amount_by_items(action, order, restriction, discount_ids, False)

            if action.name == ActionTypes.discountAmountByItemsOnTotal.value:
                result = service.apply_discount_amount_by_items(action, order, restriction, discount_ids, True)

            if action.name == ActionTypes.discountPercentageByItems.value:
                result = service.apply_discount_percentage_by_items(action, order, restriction, discount_ids, False)

            if action.name == ActionTypes.discountPercentageByGroupsOnTotal.value:
                result = service.apply_discount_percentage_by_groups(action, order, restriction, discount_ids, True)

            if action.name == ActionTypes.discountAmountByGroupsOnTotal.value:
                result = service.apply_discount_amount_by_groups(action, order, restriction, discount_ids, True)

            if action.name == ActionTypes.discountPercentageByItemsOnTotal.value:
                result = service.apply_discount_percentage_by_items(action, order, restriction, discount_ids, True)

            if action.name == ActionTypes.paymentAmount.value:
                result = service.apply_discount_amount_on_total_by_tender(action, order, restriction)

            if action.name == ActionTypes.paymentPercentage.value:
                result = service.apply_discount_percentage_on_total_by_tender(action, order, restriction)

            if action.name == ActionTypes.paymentAmountOnItem.value:
                result = service.apply_discount_amount_on_total_by_tender(action, order, restriction, True)

            if action.name == ActionTypes.itemsToAdd.value:
                result = service.apply_tender_on_added_items(action, order, uow.benefit_order_info)

            if result.benefit_applied:
                result.items_to_burn = order.get_items_to_burn(order, uow.benefit, burnt_items, self.product_groups)

            return result

        elif order.order_state == "IN_PROGRESS":
            if action.name == ActionTypes.itemsToAdd.value:
                return service.apply_items_to_add(
                    action=action,
                    burnt_items=burnt_items,
                    uow=uow,
                    restriction=restriction,
                    on_total=on_total,
                )

            if on_total is False:
                result.items_to_burn = order.get_items_to_burn(order, uow.benefit, burnt_items, self.product_groups)
                result.pending_to_apply = True
                return result

        return ActionResult([], {}, [], {}, False, False)



import unittest
import simplejson as json


class ForJson(object):
    def for_json(self):
        return {'for_json': 1}


class NestedForJson(object):
    def for_json(self):
        return {'nested': ForJson()}


class ForJsonList(object):
    def for_json(self):
        return ['list']


class DictForJson(dict):
    def for_json(self):
        return {'alpha': 1}


class ListForJson(list):
    def for_json(self):
        return ['list']


class TestForJson(unittest.TestCase):
    def assertRoundTrip(self, obj, other, for_json=True):
        if for_json is None:
            # None will use the default
            s = json.dumps(obj)
        else:
            s = json.dumps(obj, for_json=for_json)
        self.assertEqual(
            json.loads(s),
            other)

    def test_for_json_encodes_stand_alone_object(self):
        self.assertRoundTrip(
            ForJson(),
            ForJson().for_json())

    def test_for_json_encodes_object_nested_in_dict(self):
        self.assertRoundTrip(
            {'hooray': ForJson()},
            {'hooray': ForJson().for_json()})

    def test_for_json_encodes_object_nested_in_list_within_dict(self):
        self.assertRoundTrip(
            {'list': [0, ForJson(), 2, 3]},
            {'list': [0, ForJson().for_json(), 2, 3]})

    def test_for_json_encodes_object_nested_within_object(self):
        self.assertRoundTrip(
            NestedForJson(),
            {'nested': {'for_json': 1}})

    def test_for_json_encodes_list(self):
        self.assertRoundTrip(
            ForJsonList(),
            ForJsonList().for_json())

    def test_for_json_encodes_list_within_object(self):
        self.assertRoundTrip(
            {'nested': ForJsonList()},
            {'nested': ForJsonList().for_json()})

    def test_for_json_encodes_dict_subclass(self):
        self.assertRoundTrip(
            DictForJson(a=1),
            DictForJson(a=1).for_json())

    def test_for_json_encodes_list_subclass(self):
        self.assertRoundTrip(
            ListForJson(['l']),
            ListForJson(['l']).for_json())

    def test_for_json_ignored_if_not_true_with_dict_subclass(self):
        for for_json in (None, False):
            self.assertRoundTrip(
                DictForJson(a=1),
                {'a': 1},
                for_json=for_json)

    def test_for_json_ignored_if_not_true_with_list_subclass(self):
        for for_json in (None, False):
            self.assertRoundTrip(
                ListForJson(['l']),
                ['l'],
                for_json=for_json)

    def test_raises_typeerror_if_for_json_not_true_with_object(self):
        self.assertRaises(TypeError, json.dumps, ForJson())
        self.assertRaises(TypeError, json.dumps, ForJson(), for_json=False)



import sysactions

from utilconfigurations import must_require_void_reason
from utils import get_void_reason_id, get_canceled_by
from custom_sysactions import user_authorized_to_execute_action
from manager.reports import justification_report


@sysactions.action
@user_authorized_to_execute_action
def do_void_lines(pos_id, line_numbers=None, auth_data=None):
    if not line_numbers:
        return False

    model = sysactions.get_model(pos_id)
    sysactions.check_current_order(pos_id, model=model)

    if not sysactions.can_void_line(model, line_numbers):
        sysactions.show_messagebox(pos_id, "$NEED_TO_HAVE_ITEM_TO_VOID")
        return False

    auth_user = auth_data if auth_data is not None else None
    canceled_by = get_canceled_by(
        pos_id,
        model,
        auth_user
    )

    void_reason = None
    void_type = "VOID_LINE"
    if must_require_void_reason(void_type):
        void_reason = get_void_reason_id(
            pos_id=pos_id,
            void_type=void_type,
            authentication_user=canceled_by,
        )
        if void_reason is None:
            return True

    pos_ot = sysactions.get_posot(model)
    order = model.find("CurrentOrder/Order")
    order_id = order.get("orderId")
    sale_lines = order.findall("SaleLine")

    pos_ot.voidLine(int(pos_id), line_numbers)

    if void_reason:
        justification_report(
            pos_id=pos_id,
            justification_type="VOID_LINE",
            justification=void_reason.get('VOID_REASON_DESCR'),
            authorized_by=canceled_by,
            order_id=order_id
        )

    for line_number in line_numbers.split('|'):
        custom_properties = {}
        custom_properties.update(void_reason or "")
        if void_reason is None:
            custom_properties.update({"VOID_AUTH_USER": canceled_by})

        current_line_sale_lines = [x for x in sale_lines if x.get("lineNumber") == line_number]
        for current_line in current_line_sale_lines:
            item_type = current_line.get("itemType")
            if item_type == "OPTION":
                continue

            voided_qty = float(current_line.get("qty", 0))
            if item_type != "PRODUCT":
                voided_qty = voided_qty - float(current_line.get("defaultQty", 0))
            custom_properties.update({"VOID_QTY": str(voided_qty)})

            pos_ot.setOrderCustomProperties(
                custom_properties,
                orderid=order_id,
                linenumber=line_number,
                itemid=current_line.get("itemId"),
                level=current_line.get("level"),
                partcode=current_line.get("partCode"),
            )

    return True



from commons.report import Generator
from reports_app.opendaysummary.dto import OpenDaySummaryInfo  # noqa
from reports_app.opendaysummary.dto import OpenDaySummaryDto  # noqa
from reports_app.opendaysummary.dto import OpenDaySummaryHeaderDto
from reports_app.opendaysummary.dto import OpenDaySummaryBodyDto


class OpenDayGenerator(Generator):
    def __init__(self, open_day_summary_info):
        # type: (OpenDaySummaryInfo) -> ()
        self.open_day_summary_info = open_day_summary_info

    def generate_data(self):
        # type: (OpenDaySummaryInfo) -> OpenDaySummaryDto
        open_day_summary_header_dto = OpenDaySummaryHeaderDto(
            self.open_day_summary_info.pos_id,
            self.open_day_summary_info.store_id,
            self.open_day_summary_info.operator_id,
            self.open_day_summary_info.business_date
        )
        open_day_summary_body_dto = OpenDaySummaryBodyDto(
            self.open_day_summary_info.authorizer
        )

        return OpenDaySummaryDto(open_day_summary_header_dto, open_day_summary_body_dto)



import copy
import re
from operator import attrgetter

from typing import (
    Dict,
    List,
    Optional,
)
from xml.etree import ElementTree
from xml.sax.saxutils import escape

from old_helper import round_half_away_from_zero
from pos_model import (
    SaleItemType,
    SimplifiedSaleItem,
    TaxItem,
)


class SaleLineUtil(object):
    def __init__(self):
        return

    def get_product_name_from_tax_item(self, tax_item, sale_lines):
        # type: (TaxItem, List[ElementTree]) -> unicode
        sale_line = self._get_sale_line_by_tax_item(tax_item, sale_lines)

        if sale_line is None:
            raise Exception("SaleLine not found for product: {0}.{1}".format(tax_item.item_id, tax_item.part_code))

        product_name = escape(sale_line.get("productName"))

        return product_name

    def get_sale_line_by_tax_item(self, tax_item, sale_lines):
        # type: (TaxItem, List[ElementTree]) -> Optional[ElementTree]
        for sale_line in sale_lines:
            if sale_line.get("itemId") == tax_item.item_id and int(sale_line.get("partCode")) == tax_item.part_code:
                return sale_line

        raise Exception("SaleLine not found for product: {0}.{1}".format(tax_item.item_id, tax_item.part_code))


def simplify_sale_items(
    sale_items,                 # type: List[SimplifiedSaleItem]
    ignore_context=False,       # type: bool
    hide_unpriced_items=False,  # type: bool
):
    # type: (...) -> List[SimplifiedSaleItem]

    _validate_flat_list(sale_items)

    if ignore_context:
        result_sale_items = _simplify_sale_items_by_part_code(
            sale_items,
            hide_unpriced_items=hide_unpriced_items,
        )
    else:
        result_sale_items = _simplify_sale_items_by_context(
            sale_items,
            hide_unpriced_items=hide_unpriced_items,
        )

    result_sale_items.sort(key=attrgetter("original_position"))
    return result_sale_items


def _validate_flat_list(
    sale_items,  # type: List[SimplifiedSaleItem]
):
    # type: (...) -> None

    for item in sale_items:
        if item.sale_item.parent is not None or item.sale_item.sons:
            raise ValueError("Simplification requires flat SaleItem references")


def _simplify_sale_items_by_part_code(
    sale_items,                  # type: List[SimplifiedSaleItem]
    hide_unpriced_items=False,   # type: bool
):
    # type: (...) -> List[SimplifiedSaleItem]

    related_keys = {}  # type: Dict[str, List[SimplifiedSaleItem]]
    base_key = "{};item_price={!r}"

    for index, item in enumerate(copy.deepcopy(sale_items)):
        hide_item = item.sale_item.item_price == 0 and hide_unpriced_items
        if hide_item or item.sale_item.quantity == 0:
            continue

        item.key = base_key.format(
            item.sale_item.part_code,
            _get_actual_unit_price(item),
        )
        item.original_position = index
        related_keys.setdefault(item.key, []).append(item)

    result_sale_items = []  # type: List[SimplifiedSaleItem]
    for _, group in related_keys.items():
        result_sale_items.extend(_merge_sale_item_list(group))

    return result_sale_items


def _simplify_sale_items_by_context(
    sale_items,                 # type: List[SimplifiedSaleItem]
    hide_unpriced_items=False,  # type: bool
):
    # type: (...) -> List[SimplifiedSaleItem]

    refined_sale_items = _refine_sale_items(
        copy.deepcopy(sale_items),
        hide_unpriced_items=hide_unpriced_items,
    )

    if not hide_unpriced_items:
        return _merge_full_groups(refined_sale_items)

    return _merge_sale_item_list(refined_sale_items)


def _refine_sale_items(
    sale_items,                 # type: List[SimplifiedSaleItem]
    hide_unpriced_items=False,  # type: bool
):
    # type: (...) -> List[SimplifiedSaleItem]

    removal_list = []

    last_item_index = len(sale_items) - 1
    for index, item in enumerate(sale_items):
        item.original_position = index
        is_inner_cfh_item = _is_inner_cfh(item)
        has_quantity = item.sale_item.quantity > 0
        hide_item = item.sale_item.item_price == 0 and hide_unpriced_items
        if is_inner_cfh_item or not has_quantity or _is_option(item) or hide_item:
            _refine_sale_items_sons(
                item,
                sale_items[index + 1:],
            )
            removal_list.append(item)
        else:
            _update_item_properties(
                index,
                item,
                last_item_index,
                sale_items,
            )

    return _remove_items(
        removal_list,
        sale_items,
    )


def _update_item_properties(
    index,              # type: int
    item,               # type: SimplifiedSaleItem
    last_item_index,    # type: int
    sale_items,         # type: List[SimplifiedSaleItem]
):
    # type: (...) -> None

    if index < last_item_index:
        next_item = sale_items[index + 1]
        if _is_combo(item) and _is_inner_cfh(next_item):
            item.is_cfh_head = True

    if item.grouping_context is None:
        item.grouping_context = "{}.{}".format(
            item.sale_item.item_id,
            item.sale_item.part_code,
        )


def _remove_items(
    removal_list,   # type: List[SimplifiedSaleItem]
    sale_items,     # type: List[SimplifiedSaleItem]
):
    # type: (...) -> List[SimplifiedSaleItem]

    result_sale_items = []
    base_key = "{};item_price={!r};level={!r}"

    for item in sale_items:
        if item in removal_list:
            continue

        item.key = base_key.format(
            item.grouping_context,
            _get_actual_unit_price(item),
            item.sale_item.level,
        )
        result_sale_items.append(item)

    return result_sale_items


def _refine_sale_items_sons(
    parent_item,   # type: SimplifiedSaleItem
    items_subset,  # type: List[SimplifiedSaleItem]
):
    # type: (...) -> None

    part_code_pattern = re.compile(r"\.{}\.?".format(parent_item.sale_item.part_code))
    parent_context = "{}.{}".format(
        parent_item.sale_item.item_id,
        parent_item.sale_item.part_code,
    )

    for item in items_subset:
        same_order_id = item.order_id == parent_item.order_id
        same_line_number = item.sale_item.line_number == parent_item.sale_item.line_number
        starts_with_parent_item_id = item.sale_item.item_id.startswith(parent_context)
        item_is_son = starts_with_parent_item_id and same_order_id and same_line_number
        if item_is_son:
            item.sale_item.level -= 1
            item_context = item.grouping_context or "{}.{}".format(
                item.sale_item.item_id,
                item.sale_item.part_code,
            )
            item.grouping_context = re.sub(
                part_code_pattern,
                ".",
                item_context,
                count=1,
            ).rstrip(".")
        else:
            break


def _merge_full_groups(
    refined_sale_items,  # type: List[SimplifiedSaleItem],
):
    # type: (...) -> List[SimplifiedSaleItem]

    grouped_items = {}  # type: Dict[str, List[SimplifiedSaleItem]]
    related_items = []  # type: List[SimplifiedSaleItem]
    current_order_id = None
    current_line_number = None

    for index, item in enumerate(refined_sale_items):
        same_order_id = current_order_id == item.order_id
        same_line_number = current_line_number == item.sale_item.line_number
        if not same_order_id or not same_line_number:
            if related_items:
                k = '|'.join(sorted(i.key for i in related_items))
                grouped_items.setdefault(k, []).extend(related_items)

            related_items = []
            current_order_id = item.order_id
            current_line_number = item.sale_item.line_number

        related_items.append(item)

    if related_items:
        k = '|'.join(sorted(i.key for i in related_items))
        grouped_items.setdefault(k, []).extend(related_items)

    result_sale_items = []  # type: List[SimplifiedSaleItem]
    for _, items in grouped_items.items():
        result_sale_items.extend(_merge_sale_item_list(items))

    return result_sale_items


def _merge_sale_item_list(
    items,  # type: List[SimplifiedSaleItem]
):
    # type: (...) -> List[SimplifiedSaleItem]

    result_merge = []  # type: List[SimplifiedSaleItem]
    merged_items = []  # type: List[SimplifiedSaleItem]

    for index, item_ref in enumerate(items):
        for item in items[index + 1:]:
            if item_ref.key == item.key and item not in merged_items and not item.has_fraction:
                item_ref = _merge_sale_items(item_ref, item)
                merged_items.append(item)

        if item_ref not in merged_items:
            result_merge.append(item_ref)

    return result_merge


def _merge_sale_items(
    sl_a,  # type: SimplifiedSaleItem
    sl_b,  # type: SimplifiedSaleItem
):
    # type: (...) -> SimplifiedSaleItem

    sale_item = copy.deepcopy(sl_a.sale_item)
    sale_item.item_price = sl_a.sale_item.item_price + sl_b.sale_item.item_price
    sale_item.quantity = sl_a.sale_item.quantity + sl_b.sale_item.quantity
    sale_item.default_quantity = (sl_a.sale_item.default_quantity or 0) + (sl_b.sale_item.default_quantity or 0)
    multiplied_quantity = (sl_a.sale_item.multiplied_quantity or 0) + (sl_b.sale_item.multiplied_quantity or 0)
    sale_item.multiplied_quantity = multiplied_quantity
    original_item_price = sl_a.original_item_price or sl_a.sale_item.item_price
    return SimplifiedSaleItem(
        sale_item,
        order_id=None,
        key=sl_a.key,
        is_cfh_head=sl_a.is_cfh_head,
        grouping_context=sl_a.grouping_context,
        original_position=min(
            sl_a.original_position,
            sl_b.original_position,
        ),
        original_reference=sl_a.original_reference,
        original_item_price=original_item_price,
    )


def _get_actual_unit_price(
    item,  # type: SimplifiedSaleItem
):
    # type: (...) -> float

    item_price = item.original_item_price or item.sale_item.item_price
    quantity = item.sale_item.quantity
    if item.sale_item.multiplied_quantity is not None:
        quantity *= item.sale_item.multiplied_quantity

    item_price /= quantity
    return round_half_away_from_zero(item_price, precision=2)


def _is_combo(
    item,  # type: SimplifiedSaleItem
):
    # type: (...) -> bool

    item_type = item.sale_item.type or ""
    return item_type.lower() == SaleItemType.combo.lower()


def _is_inner_cfh(
    item,  # type: SimplifiedSaleItem
):
    # type: (...) -> bool

    if not _is_combo(item):
        return False

    matcher = re.match(
        r"^CFH\d+",
        item.sale_item.product_name.strip(),
        re.IGNORECASE
    )
    return matcher is not None


def _is_option(
    item,  # type: SimplifiedSaleItem
):
    # type: (...) -> bool

    item_type = item.sale_item.type or ""
    return item_type.lower() == SaleItemType.option.lower()



from abc import ABCMeta, abstractmethod

from datetime import datetime  # noqa


class SessionRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_unique_user_session_count(self, initial_date, end_date):
        # type: (datetime, datetime) -> int
        raise NotImplementedError()



"""Fixer for 'raise E, V'

From Armin Ronacher's ``python-modernize``.

raise         -> raise
raise E       -> raise E
raise E, V    -> raise E(V)

raise (((E, E'), E''), E'''), V -> raise E(V)


CAVEATS:
1) "raise E, V" will be incorrectly translated if V is an exception
   instance. The correct Python 3 idiom is

        raise E from V

   but since we can't detect instance-hood by syntax alone and since
   any client code would have to be changed as well, we don't automate
   this.
"""
# Author: Collin Winter, Armin Ronacher

# Local imports
from lib2to3 import pytree, fixer_base
from lib2to3.pgen2 import token
from lib2to3.fixer_util import Name, Call, is_tuple

class FixRaise(fixer_base.BaseFix):

    BM_compatible = True
    PATTERN = """
    raise_stmt< 'raise' exc=any [',' val=any] >
    """

    def transform(self, node, results):
        syms = self.syms

        exc = results["exc"].clone()
        if exc.type == token.STRING:
            msg = "Python 3 does not support string exceptions"
            self.cannot_convert(node, msg)
            return

        # Python 2 supports
        #  raise ((((E1, E2), E3), E4), E5), V
        # as a synonym for
        #  raise E1, V
        # Since Python 3 will not support this, we recurse down any tuple
        # literals, always taking the first element.
        if is_tuple(exc):
            while is_tuple(exc):
                # exc.children[1:-1] is the unparenthesized tuple
                # exc.children[1].children[0] is the first element of the tuple
                exc = exc.children[1].children[0].clone()
            exc.prefix = u" "

        if "val" not in results:
            # One-argument raise
            new = pytree.Node(syms.raise_stmt, [Name(u"raise"), exc])
            new.prefix = node.prefix
            return new

        val = results["val"].clone()
        if is_tuple(val):
            args = [c.clone() for c in val.children[1:-1]]
        else:
            val.prefix = u""
            args = [val]

        return pytree.Node(syms.raise_stmt,
                           [Name(u"raise"), Call(exc, args)],
                           prefix=node.prefix)



from production.box import HideOptionBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class HideOptionBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(HideOptionBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository

    def build(self, config):
        # type: (BoxConfiguration) -> HideOptionBox

        inherit_comments = self.get_extra(config, "InheritComments", "false").lower() == "true"
        return HideOptionBox(
            name=config.name,
            sons=config.sons,
            fix_default_qty=self.get_extra(config, "FixDefaultQuantity", "false").lower() == "true",
            inherit_comments=inherit_comments,
            logger=self.get_logger(config),
        )



from datetime import datetime
from logging import Logger
from threading import Lock

from messagebus import Event, MessageBus
from typing import Dict, List, Optional

from ._Subscriber import Subscriber
from .._Cleaner import Cleaner
from .._EventPublisher import EventPublisher
from .._SubscribeCreator import SubscriberCreator
from .._Waiter import Waiter
from .._WaiterRetriever import WaiterRetriever


class EventManager(SubscriberCreator, WaiterRetriever, EventPublisher, Cleaner):
    def __init__(self, message_bus, idle_timeout, bump_timeout, logger):
        # type: (MessageBus, float, float, Logger) -> None
        self.message_bus = message_bus
        self.idle_timeout = idle_timeout
        self.bump_timeout = bump_timeout
        self.logger = logger

        self.lock = Lock()
        self.subscribers = {}  # type: Dict[str, List[Subscriber]]
        self.syncs = {}  # type: Dict[str, Subscriber]

    def create_subscriber(self, subject):
        # type: (str) -> int
        must_subscribe = False
        subscriber = Subscriber()
        with self.lock:
            if subject not in self.subscribers:
                must_subscribe = True
                self.subscribers[subject] = []

            self.subscribers[subject].append(subscriber)
            self.syncs[subscriber.sync_id] = subscriber

        if must_subscribe:
            self.message_bus.subscribe(subject)

        return subscriber.sync_id

    def get_waiter(self, sync_id):
        # type: (str) -> Optional[Waiter]
        with self.lock:
            if sync_id not in self.syncs:
                return None

            subscriber = self.syncs[sync_id]
            waiter = subscriber.get_waiter()
            del self.syncs[sync_id]
            self.syncs[waiter.get_sync_id()] = subscriber

        return waiter

    def new_event(self, event):
        # type: (Event) -> None
        with self.lock:
            if event.subject in self.subscribers:
                for subscriber in self.subscribers[event.subject]:
                    subscriber.new_event(event=event)

    def clean_idle_listeners(self):
        # type: () -> None
        try:
            now = datetime.now()
            events_to_remove = {}  # type: Dict[str, List[Subscriber]]
            subscribers_to_bump = {}
            with self.lock:
                self.logger.debug("Locking for Cleaner")
                for subject in self.subscribers:
                    self.logger.debug("Subject {} has {} listeners".format(subject, len(self.subscribers[subject])))
                    for subscriber in self.subscribers[subject]:
                        if self.is_idle(now, self.idle_timeout, subscriber):
                            if subject not in events_to_remove:
                                events_to_remove[subject] = []
                            events_to_remove[subject].append(subscriber)
                        if self.need_bump(now, self.bump_timeout, subscriber):
                            if subject not in subscribers_to_bump:
                                subscribers_to_bump[subject] = []
                            subscribers_to_bump[subject].append(subscriber)

                for subject in events_to_remove:
                    for subscriber in events_to_remove[subject]:
                        self.subscribers[subject].remove(subscriber)
                        del self.syncs[subscriber.sync_id]

                for subject in subscribers_to_bump:
                    for subscriber in subscribers_to_bump[subject]:
                        event_data = "<Event subject=\"{}\" type=\"\"></Event>".format(subject)
                        subscriber.new_event(Event(subject, "", event_data))
            self.logger.debug("Releasing Lock")
        except BaseException as _:  # noqa
            self.logger.exception("Error cleaning idle listeners")

    def is_idle(self, now, timeout, subscriber):
        return subscriber.pending_waiter is None and (now - subscriber.last_seen).total_seconds() > timeout

    def need_bump(self, now, timeout, subscriber):
        return subscriber.pending_waiter is not None and (now - subscriber.last_seen).total_seconds() > timeout



import base64
import json
import time
from logging import Logger
from xml.etree import cElementTree as eTree

from model.error_type import OperationErrorsType, PosValidationErrorsType
from model.exception import OperationErrors, ValidationError
from interactor.route.payment.insert.dto import OrderInsertPaymentRequest
from model.services import FiscalServiceModel
from model.tokens import BusTokens
from interactor.route.electronic_payment.send_responses.dto import SendResponsesRequest
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, DataType
from msgbus import TK_SYS_ACK, TK_PRN_PRINT, TK_SYS_NAK
from mwhelper import BaseRepository
from persistence import Connection
from posot import OrderTaker
from sysactions import get_used_service
from typing import (
    Dict,
    List,
    Optional
)


class FiscalService(FiscalServiceModel, BaseRepository):

    def __init__(self, message_bus, logger):
        # type: (MbContextMessageBus, Logger) -> None

        super(FiscalService, self).__init__(message_bus.mbcontext)
        self.message_bus = message_bus
        self.logger = logger
        self.electronic_payments_id = None

    def process_order_fiscal(
        self,
        model,               # type: eTree
        pos_ot,              # type: OrderTaker
        print_coupon=True,   # type: bool
    ):
        # type: (...) -> Optional[str]

        max_retry = 2
        while True:
            fiscal_response = self._fiscalize_order(model)

            self.logger.info("FiscalResponse: {}".format(fiscal_response.data))

            if fiscal_response.token == TK_SYS_ACK:
                if print_coupon:
                    self._print_fiscal_coupon(model, fiscal_response)
                return fiscal_response.data

            if fiscal_response.data in ("Certificado da Loja Expirado", "Falha na Leitura da Validade do Certificado NFCE"):
                raise ValidationError(PosValidationErrorsType.FiscalCertificateError)

            if len(fiscal_response.data.split('\0')) < 2:
                is_fiscal_ok = "False"
            else:
                is_fiscal_ok, message = fiscal_response.data.split('\0')

            if is_fiscal_ok == "True":
                self.logger.info("Order fiscalized but without fiscal coupon print")
                break

            if max_retry > 0:
                max_retry -= 1
                self.logger.error("Cannot fiscalize coupon. Remaining attempts: {}".format(max_retry))
                time.sleep(1)
                continue

            raise StopIteration()

    def _print_fiscal_coupon(self, model, fiscal_response):
        # type: (eTree, Message) -> None

        try:
            print_response = self._send_print_data(model, fiscal_response)
            if print_response.token == TK_SYS_NAK:
                self.logger.error("Cannot print fiscal coupon")
        except Exception as _:
            self.logger.exception("Error printing fiscal coupon")

    def _send_print_data(self, model, response):
        # type: (eTree, Message) -> Message

        current_printer = get_used_service(model, "printer")

        message = Message(token=TK_PRN_PRINT, data_type=DataType.param, data=response.data)
        return self.message_bus.send_message(current_printer, message)

    def _fiscalize_order(self, model):
        # type: (eTree) -> Message

        order = self._get_current_order(model)
        pos_id = order.get("posId")
        order_id = order.get("orderId")
        data = "\0".join([pos_id, order_id])

        message = Message(token=BusTokens.TK_FISCALWRAPPER_PROCESS_REQUEST, data_type=DataType.param, data=data)
        return self.message_bus.send_message("FiscalWrapper", message)

    def process_payment(
        self,
        model,      # type: eTree
        request     # type: OrderInsertPaymentRequest
    ):
        # type: (...) -> Optional[str]

        self._save_promotional_tenders_payment_data(model)

        if request.electronic_type_id in self.electronic_payments_id:
            return self._process_electronic_payment(model, request)

        return None

    def _process_electronic_payment(
        self,
        model,      # type: eTree
        request     # type: OrderInsertPaymentRequest
    ):
        # type: (...) -> str

        payment_service = self._get_electronic_payment_service(model)
        try:
            response = self._add_electronic_payment(model, payment_service, request)
            self._validate_electronic_payment_response(response, request)
        except Exception as _:
            self.logger.exception("Error processing electronic payment")
            self._cancel_electronic_payment(payment_service)
            raise

        fiscal_params = self._get_electronic_fiscal_params(model, request, response)
        self._store_payment_data(fiscal_params)

        return self._get_tender_details(eTree.fromstring(response.data))

    def _add_electronic_payment(self, model, payment_service, request):
        # type: (eTree, str, OrderInsertPaymentRequest) -> Message

        data = self._get_electronic_payment_data(model, request)
        message = Message(token=BusTokens.TK_ELECTRONIC_PAYMENT_ADD_PAYMENT, data_type=DataType.param, data=data)
        return self.message_bus.send_message(payment_service, message)

    def cancel_electronic_payment(
        self,
        model,      # type: eTree
        request,    # type: OrderInsertPaymentRequest
    ):
        # type: (...) -> None

        if request.electronic_type_id in self.electronic_payments_id:
            payment_service = self._get_electronic_payment_service(model=model)
            self._cancel_electronic_payment(payment_service=payment_service)

    def _cancel_electronic_payment(self, payment_service):
        # type: (str) -> None

        message = Message(token=BusTokens.TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT)
        self.message_bus.send_message(payment_service, message)

    def _get_electronic_fiscal_params(self, model, request, response):
        # type: (eTree, OrderInsertPaymentRequest, Message) -> Dict

        order = self._get_current_order(model)
        change_amount = "0"
        tender_seq_id = str(self._get_last_tender_seq_id(order))

        fiscal_params = self._build_fiscal_params(
            order.get("posId"),
            order.get("orderId"),
            tender_seq_id,
            str(request.payment_id),
            str(request.amount),
            change_amount
        )

        return self._build_additional_fiscal_params(fiscal_params, response)

    def _validate_electronic_payment_response(self, response, request):
        # type: (Message, OrderInsertPaymentRequest) -> None

        if response.token != TK_SYS_ACK or not response.data:
            raise OperationErrors(OperationErrorsType.InvalidPaymentResponse)

        response_xml = eTree.fromstring(response.data)
        result = response_xml.get("Result")
        if result != "0":
            raise OperationErrors(OperationErrorsType.UnexpectedPaymentResponse)

        approved_amount = response_xml.get("ApprovedAmount").replace(",", ".")
        if not approved_amount or float(approved_amount) == float(request.amount):
            return

        message = "Approved amount [{}] is different of the order amount [{}]".format(approved_amount, request.amount)
        self.logger.info(message)
        request.amount = approved_amount

    @staticmethod
    def _get_electronic_payment_service(model):
        # type: (eTree) -> str

        payment_service = get_used_service(model, "eft")
        if payment_service:
            return payment_service
        raise OperationErrors(OperationErrorsType.InvalidPaymentService)

    def _get_electronic_payment_data(self, model, request):
        # type: (eTree, OrderInsertPaymentRequest) -> str

        order = self._get_current_order(model)

        pos_id = str(request.pos_id)
        order_id = order.get("orderId")
        session_id = model.find("Operator").get("sessionId")
        operator = model.find("Operator").get("id")
        amount = "{:.2f}".format(request.amount)
        electronic_type = str(request.electronic_type_id)
        tender_seq_id = ''

        totaled_states = order.findall("StateHistory/State[@state='TOTALED']")
        last_totaled = totaled_states[-1].get("timestamp").replace("-", "").replace(":", "")
        last_totaled_date = last_totaled
        fiscal_date = last_totaled_date[:8]
        fiscal_time = last_totaled_date[9:15]

        data = ";".join(
            [
                pos_id,
                order_id,
                session_id,
                operator,
                electronic_type,
                amount,
                fiscal_date,
                fiscal_time,
                tender_seq_id
            ]
        )

        return data

    def _save_promotional_tenders_payment_data(self, model):
        # type: (eTree) -> None

        order = self._get_current_order(model)
        promotional_tenders = self._get_order_promotional_tenders(order)
        for tender in promotional_tenders:
            fiscal_params = self._build_fiscal_params(order.get("posId"),
                                                      order.get("orderId"),
                                                      promotional_tenders[tender].get("tender_seq_id"),
                                                      promotional_tenders[tender].get("type"),
                                                      promotional_tenders[tender].get("amount"),
                                                      promotional_tenders[tender].get("change"))

            self._store_payment_data(fiscal_params)

    @staticmethod
    def _build_fiscal_params(pos_id, order_id, tender_seq_id, tender_type, amount, change):
        # type: (str, str, str, str, str, str) -> Dict

        return {
            "pos_id": pos_id,
            "order_id": order_id,
            "tender_seq_id": tender_seq_id,
            "type": tender_type,
            "amount": amount,
            "change": change
        }

    @staticmethod
    def _build_additional_fiscal_params(fiscal_params, response):
        # type: (Dict, Message) -> Dict

        response_xml = eTree.fromstring(response.data)
        if response_xml is not None:
            fiscal_params["auth_code"] = response_xml.get("AuthCode", "")
            fiscal_params["id_auth"] = response_xml.get("IdAuth", "")
            fiscal_params["cnpj_auth"] = response_xml.get("CNPJAuth", "")
            fiscal_params["bandeira"] = response_xml.get("Bandeira", "")
            fiscal_params["receipt_merchant"] = base64.b64decode(response_xml.get("ReceiptMerchant", ""))
            fiscal_params["receipt_customer"] = base64.b64decode(response_xml.get("ReceiptCustomer", ""))

        return fiscal_params

    @staticmethod
    def _get_order_promotional_tenders(order):
        # type: (eTree) -> Dict

        promotional_tenders = {}
        for tender in order.findall("TenderHistory/Tender"):
            tender_detail = tender.get("tenderDetail")
            if tender_detail:
                is_promotional_tender = json.loads(tender_detail).get("promotionalTender", False)
                if is_promotional_tender:
                    tender_id = tender.get("tenderId")
                    promotional_tenders[tender_id] = {"amount": tender.get("tenderAmount"),
                                                      "tender_seq_id": tender_id,
                                                      "type": tender.get("tenderType"),
                                                      "change": "0"}
        return promotional_tenders

    def _store_payment_data(self, fiscal_params):
        # type: (Dict) -> None

        def inner_func(conn):
            # type: (Connection) -> None

            conn.pquery("fiscal_savePaymentData", **fiscal_params)

        self.execute_with_transaction(inner_func, service_name="FiscalPersistence")

    def get_electronic_payments_id(self):
        # type: () -> None

        def inner_func(conn):
            # type: (Connection) -> List[int]

            query = "SELECT ElectronicType FROM TenderType WHERE ElectronicType <> 0"
            cursor = conn.select(query)

            electronic_type_ids = []
            for row in cursor:
                electronic_type_id = int(row.get_entry("ElectronicType"))
                if electronic_type_id not in electronic_type_ids:
                    electronic_type_ids.append(electronic_type_id)

            return electronic_type_ids

        self.electronic_payments_id = self.execute_with_connection(inner_func, service_name="Persistence")

    def send_electronic_payment_message_response(self, model, request):
        # type: (eTree, SendResponsesRequest) -> None

        payment_service = self._get_electronic_payment_service(model)
        data = "|".join([model.get("posId"), request.response])
        message = Message(token=BusTokens.TK_ELECTRONIC_PAYMENT_SEND_RESPONSE, data_type=DataType.param, data=data)
        response = self.message_bus.send_message(payment_service, message)
        if response.token != TK_SYS_ACK:
            raise OperationErrors(OperationErrorsType.InvalidPaymentResponse)

    @staticmethod
    def _get_last_tender_seq_id(order):
        return len(order.findall("TenderHistory/Tender")) + 1

    @staticmethod
    def _get_current_order(model):
        return model.find("CurrentOrder/Order")

    @staticmethod
    def _get_tender_details(
        payment_data  # type: Dict
    ):
        # type: (...) -> str

        tender_details = json.dumps({
            "Media": payment_data.get("Media", "").strip(),
            "CNPJAuth": payment_data.get("CNPJAuth", ""),
            "TransactionProcessor": payment_data.get("TransactionProcessor", ""),
            "Bandeira": payment_data.get("Bandeira", ""),
            "IdAuth": payment_data.get("IdAuth", ""),
            "AuthCode": payment_data.get("AuthCode", ""),
            "ReceiptCustomer": payment_data.get("ReceiptCustomer", ""),
            "ReceiptMerchant": payment_data.get("ReceiptMerchant", ""),
            "AccreditorSat": payment_data.get("AccreditorSat", ""),
            "NfceBrand": payment_data.get("NfceBrand", ""),
            "NsuHost": payment_data.get("NsuHost", ""),
            "CardNumber": payment_data.get("CardNumber", ""),
            "TransactionIdentifier": payment_data.get("TransactionIdentifier"),
            "OriginalNfceBrand": payment_data.get("OriginalNfceBrand", "")
        })

        return tender_details

    def get_customer_receipts(
        self,
        model,      # type: eTree
    ):
        # type: (...) -> List[str]

        order = self._get_current_order(model)
        customer_receipts = []
        for tender in order.findall("TenderHistory/Tender"):
            if int(tender.get("tenderType", -1)) in self.electronic_payments_id:
                tender_detail = json.loads(tender.get("tenderDetail", "{}"))
                tef_customer_receipt = tender_detail.get("ReceiptCustomer", '') if tender_detail is not None else None
                if tef_customer_receipt not in ['None', None, '']:
                    customer_receipts.append(tef_customer_receipt)

        return customer_receipts



from abc import ABCMeta, abstractmethod
from typing import List, Dict  # noqa


class Translator(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def translate_labels(self, labels, pos_id):
        # type: (List[str], int) -> Dict[str, str]
        raise NotImplementedError()



"""ISO 8601 date time string parsing

Basic usage:
>>> import iso8601
>>> iso8601.parse_date("2007-01-25T12:00:00Z")
datetime.datetime(2007, 1, 25, 12, 0, tzinfo=<iso8601.Utc ...>)
>>>

"""

import datetime
from decimal import Decimal
import sys
import re

__all__ = ["parse_date", "ParseError", "UTC",
           "FixedOffset"]

if sys.version_info >= (3, 0, 0):
    _basestring = str
else:
    _basestring = basestring


# Adapted from http://delete.me.uk/2005/03/iso8601.html
ISO8601_REGEX = re.compile(
    r"""
    (?P<year>[0-9]{4})
    (
        (
            (-(?P<monthdash>[0-9]{1,2}))
            |
            (?P<month>[0-9]{2})
            (?!$)  # Don't allow YYYYMM
        )
        (
            (
                (-(?P<daydash>[0-9]{1,2}))
                |
                (?P<day>[0-9]{2})
            )
            (
                (
                    (?P<separator>[ T])
                    (?P<hour>[0-9]{2})
                    (:{0,1}(?P<minute>[0-9]{2})){0,1}
                    (
                        :{0,1}(?P<second>[0-9]{1,2})
                        ([.,](?P<second_fraction>[0-9]+)){0,1}
                    ){0,1}
                    (?P<timezone>
                        Z
                        |
                        (
                            (?P<tz_sign>[-+])
                            (?P<tz_hour>[0-9]{2})
                            :{0,1}
                            (?P<tz_minute>[0-9]{2}){0,1}
                        )
                    ){0,1}
                ){0,1}
            )
        ){0,1}  # YYYY-MM
    ){0,1}  # YYYY only
    $
    """,
    re.VERBOSE
)

class ParseError(Exception):
    """Raised when there is a problem parsing a date string"""

if sys.version_info >= (3, 2, 0):
    UTC = datetime.timezone.utc
    def FixedOffset(offset_hours, offset_minutes, name):
        return datetime.timezone(
            datetime.timedelta(
                hours=offset_hours, minutes=offset_minutes),
            name)
else:
    # Yoinked from python docs
    ZERO = datetime.timedelta(0)
    class Utc(datetime.tzinfo):
        """UTC Timezone

        """
        def utcoffset(self, dt):
            return ZERO

        def tzname(self, dt):
            return "UTC"

        def dst(self, dt):
            return ZERO

        def __repr__(self):
            return "<iso8601.Utc>"

    UTC = Utc()

    class FixedOffset(datetime.tzinfo):
        """Fixed offset in hours and minutes from UTC

        """
        def __init__(self, offset_hours, offset_minutes, name):
            self.__offset_hours = offset_hours  # Keep for later __getinitargs__
            self.__offset_minutes = offset_minutes  # Keep for later __getinitargs__
            self.__offset = datetime.timedelta(
                hours=offset_hours, minutes=offset_minutes)
            self.__name = name

        def __eq__(self, other):
            if isinstance(other, FixedOffset):
                return (
                    (other.__offset == self.__offset)
                    and
                    (other.__name == self.__name)
                )
            return NotImplemented

        def __getinitargs__(self):
            return (self.__offset_hours, self.__offset_minutes, self.__name)

        def utcoffset(self, dt):
            return self.__offset

        def tzname(self, dt):
            return self.__name

        def dst(self, dt):
            return ZERO

        def __repr__(self):
            return "<FixedOffset %r %r>" % (self.__name, self.__offset)


def to_int(d, key, default_to_zero=False, default=None, required=True):
    """Pull a value from the dict and convert to int

    :param default_to_zero: If the value is None or empty, treat it as zero
    :param default: If the value is missing in the dict use this default

    """
    value = d.get(key) or default
    if (value in ["", None]) and default_to_zero:
        return 0
    if value is None:
        if required:
            raise ParseError("Unable to read %s from %s" % (key, d))
    else:
        return int(value)

def parse_timezone(matches, default_timezone=UTC):
    """Parses ISO 8601 time zone specs into tzinfo offsets

    """

    if matches["timezone"] == "Z":
        return UTC
    # This isn't strictly correct, but it's common to encounter dates without
    # timezones so I'll assume the default (which defaults to UTC).
    # Addresses issue 4.
    if matches["timezone"] is None:
        return default_timezone
    sign = matches["tz_sign"]
    hours = to_int(matches, "tz_hour")
    minutes = to_int(matches, "tz_minute", default_to_zero=True)
    description = "%s%02d:%02d" % (sign, hours, minutes)
    if sign == "-":
        hours = -hours
        minutes = -minutes
    return FixedOffset(hours, minutes, description)

def parse_date(datestring, default_timezone=UTC):
    """Parses ISO 8601 dates into datetime objects

    The timezone is parsed from the date string. However it is quite common to
    have dates without a timezone (not strictly correct). In this case the
    default timezone specified in default_timezone is used. This is UTC by
    default.

    :param datestring: The date to parse as a string
    :param default_timezone: A datetime tzinfo instance to use when no timezone
                             is specified in the datestring. If this is set to
                             None then a naive datetime object is returned.
    :returns: A datetime.datetime instance
    :raises: ParseError when there is a problem parsing the date or
             constructing the datetime instance.

    """
    if not isinstance(datestring, _basestring):
        raise ParseError("Expecting a string %r" % datestring)
    m = ISO8601_REGEX.match(datestring)
    if not m:
        raise ParseError("Unable to parse date string %r" % datestring)
    groups = m.groupdict()

    tz = parse_timezone(groups, default_timezone=default_timezone)

    groups["second_fraction"] = int(Decimal("0.%s" % (groups["second_fraction"] or 0)) * Decimal("1000000.0"))

    try:
        return datetime.datetime(
            year=to_int(groups, "year"),
            month=to_int(groups, "month", default=to_int(groups, "monthdash", required=False, default=1)),
            day=to_int(groups, "day", default=to_int(groups, "daydash", required=False, default=1)),
            hour=to_int(groups, "hour", default_to_zero=True),
            minute=to_int(groups, "minute", default_to_zero=True),
            second=to_int(groups, "second", default_to_zero=True),
            microsecond=groups["second_fraction"],
            tzinfo=tz,
        )
    except Exception as e:
        raise ParseError(e)



import json
from typing import Optional

from report import Generator
from reportsdata.justificationreport.dto import (
    JustificationGeneratorDto,
    JustificationHeaderDto,
    JustificationBodyDto,
)

from utils.report_formatters import remove_accents


class JustificationReportGenerator(Generator):

    def __init__(
        self,
        pos_id,     # type: str
        data,       # type: str
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.data = json.loads(
            s=data,
        )

    def generate_data(
        self
    ):
        # type: (...) -> JustificationGeneratorDto

        return JustificationGeneratorDto(
            header=self._get_header(),
            body=self._get_body(),
        )

    def _get_header(
        self
    ):
        # type: (...) -> JustificationHeaderDto

        justification_type = self._get_data(
            key="justificationType",
        )

        title = '$JUSTIFICATION_DISCOUNT_TITLE'
        subtitle = ""
        if justification_type != 'APPLY_DISCOUNT':
            title = '$JUSTIFICATION_CANCELLATION_TITLE'
            subtitle = '$JUSTIFICATION_TITLE_ORDER'
            if str(justification_type) == "VOID_LINE":
                subtitle = '$JUSTIFICATION_TITLE_ITEM'

        return JustificationHeaderDto(
            title=title,
            subtitle=subtitle,
        )

    def _get_body(
        self
    ):
        # type: (...) -> JustificationBodyDto

        order_id = self._get_data(
            key="orderId",
        )

        operator = self._get_data(
            key="operator",
        )

        authorized_by = self._get_data(
            key="AuthorizedBy",
        )

        discount_description = remove_accents(
            text=self._get_data(
                key="discountUserDescription",
            )
        )

        justification = remove_accents(
            text=self._get_data(
                key="justification",
            )
        )

        return JustificationBodyDto(
            order_id=order_id,
            operator=operator,
            authorized_by=authorized_by,
            discount_description=discount_description,
            justification=justification,
        )

    def _get_data(
        self,
        key,                # type: str
        default="",         # type: Optional[str]
    ):
        # type: (...) -> str

        return self.data.get(
            key,
            default,
        )



from ._ApplyBenefitProcessor import ApplyBenefitProcessor
from ._GetBenefitInfoProcessor import GetBenefitInfoProcessor
from ._OrderModifiedReopenedProcessor import OrderModifiedReopenedProcessor
from ._OrderModifiedPaidProcessor import OrderModifiedPaidProcessor
from ._OrderModifiedVoidProcessor import OrderModifiedVoidProcessor
from ._OrderModifiedVoidLineProcessor import OrderModifiedVoidLineProcessor
from ._NotMappedProcessor import NotMappedProcessor
from ._TestConnectionProcessor import TestConnectionProcessor



from typing import Dict, Any
from pos_model import Order
from _SatXmlPartBuilder import SatXmlPartBuilder


class TotalBuilder(SatXmlPartBuilder):
    def __init__(self, tax_builder):
        self.tax_builder = tax_builder

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        total_taxes = self.tax_builder.ibpt_tax_calculator(order.order_picture_xml)
        total_tax_amount = float(total_taxes['nacionalfederal']['value']) + float(total_taxes['estadual']['value'])
        return "<total><vCFeLei12741>{0:.02f}</vCFeLei12741></total>".format(total_tax_amount)
